import sys
import numpy as np
import scipy as sp
from scipy.special import factorial
from scipy.interpolate import RegularGridInterpolator
import itertools
import os.path
import h5py
from functools import partial
import multiprocessing

if 'AHFpy' in sys.modules:
    from AHFpy.data import t_ij, t_ij_k, sym2, sym3
    from AHFpy.data import Interp3D, Timers
    from AHFpy.data import read_h5_dset, read_h5_dset_athena
    from AHFpy.Puncture import Puncture
else:
    # Assume we are in ahfpy/AHFpy/
    from data import t_ij, t_ij_k, sym2, sym3
    from data import Interp3D, Timers
    from data import read_h5_dset, read_h5_dset_athena
    from Puncture import Puncture

class AHF():
    r"""
    AHF: Apparent horizon finder 

    This code is shaped from the BAM's AHmod (Brugmann, Gonzalez, Lages).
    The implemented method follows Gundlach \cite{Gundlach:1997us} and
    Alcubierre+ \cite{Alcubierre:1998rq}.   

    AHF calculates the AHF as a height function around a central
    point. The height function is given by values $h(\theta,\phi)$ at
    discrete coordinate angles or by the coefficients of the spherical
    harmonics expansion
    $$
    h(\theta,\phi) = \sqrt{4\pi} \sum_{l=0}^{l_{max}} 
                     \sum_{m=-l}^{l} a_{lm} Y_{lm}(\theta,\phi).
    $$
    In \cite{Alcubierre:1998rq} the overall factor $\sqrt{4\pi}$ is explained 
    as then $a_{00}$ is the average coordinate radius, $a_{10}$ is its average 
    displacement in the $z$-direction and so on.

    The shape of the apparent horizon is governed by the expansion
    $$
    H = \nabla_i s^i + K_{ij}\,s^i s^j - K.
    $$
    Here $s^i$ is the (space-like) outward normal on this surface,
    $K_{ij}$ is the extrinsic curvature and $K$ is its trace. 
    The apparent horizon is the outermost 2-surface with $H=0$
    everywhere. 
    
    The nonlinear elliptic equation is solved with the 'fast flow algorithm':
    Starting from a sphere that is definitely outside the horizon it
    flows in pseudo time against the solution.  
    """
    
    def __init__(self,
                 alpha_beta_const = 1, # constant to determine alpha & beta fast flow pars
                 flow_iter = 100, # Max iterations for the fast flow
                 # ### OLD setup for evenly spaced grid:
                 #nphi = 30, # Number of points in phi direction
                 #ntheta = 60, # Number of circles in theta direction
                 #glgrid = False, # Use Gauss-Legendre nodes and GL quadratures
                 # ### NEW setup:
                 nphi = 16, # Number of points in phi direction
                 ntheta = 8, # Number of circles in theta direction
                 glgrid = True, # Use Gauss-Legendre nodes and GL quadratures
                 #
                 lmax = 8, # Maximum number of Legendre polynomials
                 mass_tol = 1e-3, # Mass tolerance, stopping criterion for flow
                 hmean_tol = 100., # h-mean tolerance
                 time = (0, 1), # Time window to search
                 nhorizons = 1, # Such many surfaces (MOTS) will be searched 
                 use_last = True, # Use the last surface found as initial guess
                 initial_guess_expand = 1.1, # Factor to make initial guess larger
                 initial_radius = [1.0], # List of guesses, one per horizon
                 central_points = [(0.,0.,0.)], # list of central points
                 interpolation_method = 'linear', # Order of interpolation onto surface  
                 compute_metric_drvts = True, # Compute first derivative of the 3-metric
                 output = True, # Output results about MOTS
                 output_modes = False, # Output spherical modes of MOTS
                 output_iterations = False, # Output file with time and fast-flow iteration
                 output_surface = False, # Output file with horizon shape
                 outputdir = '.',
                 # ... 
                 #output_xyt = False, #TODO Output location points of MOTS
                 #mpiproc = 0, #TODO MPI processors (for MPI run) 
                 verbose = True,
                 timers_on = True):

        self.flow_iter = flow_iter
        self.alpha_beta_const = alpha_beta_const

        if lmax<1:
            raise ValueError("lmax must be >0")
        self.lmax = lmax

        self.nhorizons = nhorizons
        self.use_last = use_last
        self.initial_guess_expand = initial_guess_expand
        self.initial_radius = initial_radius
        self.central_points = central_points

        self.mass_tol = mass_tol
        self.hmean_tol = hmean_tol
        
        self.compute_metric_drvts = compute_metric_drvts
        self.interpolation_method = interpolation_method

        self.tbeg = time[0]
        self.tend = time[1]
        
        self.verbose = verbose
        self.output = output
        self.output_modes = output_modes
        self.output_iterations = output_iterations
        self.output_surface = output_surface
        
        self.outputdir = outputdir
        os.makedirs(outputdir, exist_ok=True)

        # Interpolator
        self.Int3 = self._Int3_LinCustom
        #self.Int3 = self._Int3_LinRGI
        
        self.timers_on = timers_on
        if self.timers_on:
            self.timer = Timers(outputdir = outputdir)
            self.timer.start('init')
        
        # Mesh grid in (theta,phi) and integration weights
        self.ntheta = ntheta
        self.nphi = nphi
        if self.nphi < 1:
            raise ValueError("nphi cannot be < 1")
        if self.ntheta < 1:
            raise ValueError("ntheta cannot be < 1")
        self.glgrid = glgrid
        self.th, self.ph, self.weights = np.array([]), np.array([]), np.array([])
        self.th, self.ph, self.weights = self.TPMesh()
        
        self.Y,self.dY_dth, self.dY_dph, self.d2Y_dth2, self.d2Y_dph2, self.d2Y_dthdph = self.SphericalHarmonics() 

        self.ahfound = np.full(nhorizons, False, dtype=bool)
        self.last_rh = np.zeros(nhorizons)

        if self.timers_on: self.timer.stop('init')
        
    def type(self):
        """
        Returns the class type
        """
        return type(self)

    def dtheta(self):
        """
        theta grid spacing
        """
        return np.pi/self.ntheta

    def theta(self):
        """ 
        theta grid 
        """
        i = np.arange(self.ntheta)
        return self.dtheta()*(0.5 + i)
    
    def dphi(self):
        """
        phi grid spacing
        """
        return 2.0*np.pi/self.nphi

    def phi(self):
        """
        phi grid
        """
        j = np.arange(self.nphi)
        return self.dphi()*(0.5 + j)

    def grid_tp(self):
        """
        Evenly spaced latitude-longitude grid and weights
        """
        th, ph = np.meshgrid(self.theta(), self.phi(),
                             indexing = "ij")
        dphi = self.dphi()
        dtheta = self.dtheta()
        weights = np.sin(th) * dtheta * dphi
        return th, ph, weights

    def grid_gl(self, check_ntheta=True):
        """
        Gauss-Legendre nodes and weights
        """
        if self.nphi % 2:
            raise ValueError("nphi must be even")
        m = self.nphi//2
        if check_ntheta:
            if self.ntheta != m:
                raise ValueError("ntheta must be nphi/2")
            self.ntheta = m
        else:
            m = self.ntheta
        x, w = np.polynomial.legendre.leggauss(m)
        theta = sorted(np.arccos(x))
        phi = self.phi()
        th, ph = np.meshgrid(theta, phi, indexing = "ij")
        weights = np.pi/m * np.repeat(w[None],self.nphi,axis=0).T \
            * np.ones((self.ntheta,self.nphi))
        return th, ph, weights
    
    def TPMesh(self):
        """
        theta-phi mesh grid
        """
        if self.th.size == 0 or self.ph.size == 0:
            if self.glgrid:
                return self.grid_gl()
            else:
                return self.grid_tp()
        return self.th, self.ph, self.weights

    def DerivativesXYZvsTP(self,rh,drh_dth,drh_dph,FDGradient=False):
        """
        Derivatives of (x,y,z) w.r.t. (theta,phi)
        """
        if self.timers_on: self.timer.start('DerivativesXYZvsTP')
        
        th, ph, _ = self.TPMesh()
        costh = np.cos(th)
        sinth = np.sin(th)
        cosph = np.cos(ph)
        sinph = np.sin(ph)

        if FDGradient:
            drdt, drdp = np.gradient(rh, th[:,0], ph[0,:], edge_order = 1)
            #drdt, drdp = np.gradient(rh, self.dtheta(), self.dphi(), edge_order = 1)
        else:
            drdt, drdp = drh_dth, drh_dph
        
        # Derivatives of (x,y,z) w.r.t. theta
        dxdt = (drdt*sinth + rh *costh)*cosph
        dydt = (drdt*sinth + rh *costh)*sinph
        dzdt =  drdt*costh - rh *sinth
        
        # Derivatives of (x,y,z) w.r.t. phi
        dxdp = (drdp*cosph - rh *sinph)*sinth
        dydp = (drdp*sinph + rh *cosph)*sinth
        dzdp =  drdp*costh

        if self.timers_on: self.timer.stop('DerivativesXYZvsTP')
        return dxdt, dydt, dzdt, dxdp, dydp, dzdp
    
    def Legendre(self):
        """
        Legendre polys and drvts
        """
        if self.timers_on: self.timer.start('Legendre')

        th, ph, _ = self.TPMesh()        
        costh = np.cos(th)
        sinth = np.sin(th)

        # coefficients
        lmax = self.lmax
        fac = factorial(np.arange(2*lmax+2))
        l_index = np.arange(0,lmax+1)
        pow2l = np.power(2,l_index)
        sqrt3 = np.sqrt(3.)
        
        # diagonal coefs
        d = np.sqrt((2*l_index+1)*fac[2*l_index]/(4.0*np.pi))/(pow2l*fac[l_index])
        sqrt2lm1 = np.array([0, 0] + [np.sqrt(2*l-1) for l in range(2,lmax+1)])
        
        # off-diagonal coefs 
        c = {}
        csqrtlm = {}
        for l in range(2,lmax+1):
            for m in range(0,l):
                c[(l,m)] = np.sqrt((2*l+1)/(l*l-m*m))
                csqrtlm[(l,m)] = np.sqrt(((l-1)*(l-1)-m*m)/(2*l-3))

        # zeros P and drvts
        P = {} 
        dP_dth = {} 
        d2P_dth2 = {}
        for l in range(0,lmax+1):
            for m in range(0,l+1):
                P[(l,m)] = 0.0
                dP_dth[(l,m)] = 0.0
                d2P_dth2[(l,m)] = 0.0

        # Legendre functions
        # - treat first diagonal terms
        # - special treatment for all P[l][l-1] where P[l-2][l-1] is not needed
        for l in range(0,lmax+1):
            P[(l,l)] = d[l] * np.power((-sinth),l)
        P[(1,0)] = sqrt3*costh*P[(0,0)]
        for l in range(2,lmax+1):
            for m in range(0,l-1):
                P[(l,m)] = c[(l,m)] *(sqrt2lm1[l]*costh*P[(l-1,m)] - csqrtlm[(l,m)]*P[(l-2,m)])
            m = l-1
            P[(l,l-1)] = c[(l,m)] *(sqrt2lm1[l]*costh*P[(l-1,m)])
  
        # First derivatives
        for l in range(0,lmax+1):
            dP_dth[(l,l)] = d[l] *l*np.power((-sinth),l-1)*(-costh)
        dP_dth[(1,0)] = sqrt3*(-sinth*P[(0,0)] +costh*dP_dth[(0,0)])
        for l in range(2,lmax+1):
            for m in range(0,l-1):
                dP_dth[(l,m)] = c[(l,m)] *(sqrt2lm1[l]*(-sinth*P[(l-1,m)] + costh*dP_dth[(l-1,m)])-csqrtlm[(l,m)]*dP_dth[(l-2,m)])
            m = l-1
            dP_dth[(l,l-1)] = c[(l,m)] *(sqrt2lm1[l]*(-sinth*P[(l-1,m)] + costh*dP_dth[(l-1,m)]))

        # Second derivatives 
        for l in range(0,lmax+1):
            d2P_dth2[(l,l)] = d[l] *l*((l-1)*np.power(-sinth,l-2)*costh*costh + np.power(-sinth,l-1)*sinth)
        d2P_dth2[(1,0)] = sqrt3*(-costh*P[(0,0)] -2.0*sinth*dP_dth[(0,0)] +costh*d2P_dth2[(0,0)])
        for l in range(2,lmax+1):
            for m in range(0,l-1):
                d2P_dth2[(l,m)] = c[(l,m)]\
                                  *(sqrt2lm1[l]*(-costh*P[(l-1,m)] - 2.0*sinth*dP_dth[(l-1,m)] + costh*d2P_dth2[(l-1,m)])\
                                    - csqrtlm[(l,m)]*d2P_dth2[(l-2,m)])
            m = l-1
            d2P_dth2[(l,m)] = c[(l,m)]\
                              *(sqrt2lm1[l]*(-costh*P[(l-1,m)] - 2.0*sinth*dP_dth[(l-1,m)] + costh*d2P_dth2[(l-1,m)]))

        if self.timers_on: self.timer.stop('Legendre')
        return P, dP_dth, d2P_dth2

    def SphericalHarmonics(self):
        """
        Spherical Harmonics and drvts
        """
        if self.timers_on: self.timer.start('SphericalHarmonics')
        
        lmax = self.lmax
        th, ph, _ = self.TPMesh()
        P, dP_dth, d2P_dth2 = self.Legendre()

        # coefficients
        sqrt2 = np.sqrt(2.)
        exp_ph = {}
        exp_ph_d = {}
        exp_ph_d2 = {}
        for m in range(0,lmax+1):
            exp_ph[m] = np.exp(1j * m * ph)
            exp_ph_d[m] = 1j * m * exp_ph[m]
            exp_ph_d2[m] = - m**2 * exp_ph[m]

        # zeros Y and drvts
        Y = {}
        dY_dth = {}
        dY_dph = {}
        d2Y_dth2 = {}
        d2Y_dph2 = {}
        d2Y_dthdph = {}
        for l in range(0,lmax+1):
            for m in range(0,l+1):
                Y[(l,m)] = 0.0
                dY_dth[(l,m)] = 0.0
                dY_dph[(l,m)] = 0.0
                d2Y_dth2[(l,m)] = 0.0
                d2Y_dph2[(l,m)] = 0.0
                d2Y_dthdph[(l,m)] = 0.0
                
        # Spherical Harmonics
        for l in range(0,lmax+1):
            Y[(l,0)] = P[(l,0)]
        for l in range(1,lmax+1):
            for m in range(1,l+1):
                Y[(l,m)] = sqrt2 * P[(l,m)] * exp_ph[m]

        # First derivatives 
        for l in range(0,lmax+1):
            dY_dth[(l,0)] = dP_dth[(l,0)]
        for l in range(1,lmax+1):
            for m in range(1,l+1):
                dY_dth[(l,m)] = sqrt2 * dP_dth[(l,m)] * exp_ph[m]
                dY_dph[(l,m)] = sqrt2 * P[(l,m)] * exp_ph_d[m]

        # Second derivatives 
        for l in range(0,lmax+1):
            d2Y_dth2[(l,0)] = d2P_dth2[(l,0)]
        for l in range(1,lmax+1):
            for m in range(1,l+1):
                d2Y_dth2[(l,m)] = sqrt2 * d2P_dth2[(l,m)] * exp_ph[m]
                d2Y_dph2[(l,m)] = sqrt2 * P[(l,m)] * exp_ph_d2[m]
                d2Y_dthdph[(l,m)] = sqrt2 * dP_dth[(l,m)] * exp_ph_d[m]

        if self.timers_on: self.timer.stop('SphericalHarmonics')
        return Y, dY_dth, dY_dph, d2Y_dth2, d2Y_dph2, d2Y_dthdph

    def RadiusFromSphericalHarmonics(self, aa):
        """
        Radius of the surface r = a_lm Y_lm
        """
        lmax = self.lmax
        Y, dY_dth, dY_dph = self.Y, self.dY_dth, self.dY_dph
        rh = 0.0
        drh_dth, drh_dph = 0.0, 0.0
        for l in range(0,lmax+1):
            rh += aa[(l,0)] * Y[(l,0)]
            drh_dth += aa[(l,0)] * dY_dth[(l,0)]
            #drh_dph += aa[(l,0)] * dY_dph[(l,0)]
            
        for l in range(1,lmax+1):
            for m in range(1,l+1):
                dYcdtheta = dY_dth[(l,m)].real
                dYsdtheta = dY_dth[(l,m)].imag
                dYcdphi = dY_dph[(l,m)].real 
                dYsdphi = dY_dph[(l,m)].imag 

                ar = aa[(l,m)].real # = ac
                ai = aa[(l,m)].imag # = as
                
                #rh += Y[(l,m)].real*aa[(l,m)].real + Y[(l,m)].imag*aa[(l,m)].imag
                rh += ar * Y[(l,m)].real + ai * Y[(l,m)].imag
                drh_dth += ar * dYcdtheta + ai * dYsdtheta
                drh_dph += ar * dYcdphi + ai * dYsdphi
                                
        return rh, drh_dth, drh_dph
    
    def CentralPoint(self,number):
        """
        Pick a central point from a list
        """
        if number > len(self.central_points)-1:
            point = 0,0,0
        else:
            point = self.central_points[number]
        return point

    def InitialGuess(self,number):
        """
        Initial guess for the spectral coefficients $a_{lm}$.
        If a previous file exists it is read and the initial surface 
        is set to these values. 
        Otherwise get an estimate of a(l=0) from input. 
        """
        nmodes = 0
        for l in range(0,self.lmax+1):
            for m in range(1,l+1):
                nmodes += 1
        aa = {}
        fname = "{}/horizon{:02d}_modes.txt".format(self.outputdir,number)
        if os.path.isfile(fname):
            aa_prev = np.loadtxt(fname,dtype = np.complex128)
            if nmodes != len(aa_prev): 
                print("Inconsistent number of coefficients: {} != {}".format(nmodes,len(aa_prev)))
                raise
            k = 0
            for l in range(0,self.lmax+1):
                for m in range(1,l+1):
                    aa[(l,m)] = aa_prev[k]
            aa[(0,0)] *= self.initial_guess_expand
        else:
            lmax = self.lmax
            for l in range(0,lmax+1):
                for m in range(0,l+1):
                    aa[(l,m)] = 0.
            aa[(0,0)] = np.sqrt(4.0*np.pi) * self.initial_radius[number] 
        return aa

    def UpdateFlowSpectralComponents(self, aa, rho,
                                     alpha_beta_const = 1):
        """
        Find new spectral components with fast flow
        """
        if self.timers_on: self.timer.start('UpdateFlowSpectralComponents')
        
        th, ph, weights = self.TPMesh()        
        lmax = self.lmax
        Y = self.Y

        if type(alpha_beta_const) in [list,tuple] and len(alpha_beta_const):
            # in case one knows what one's doing:
            alpha = alpha_beta_const[0]
            beta = alpha_beta_const[1]
        else:
            alpha = alpha_beta_const
            beta = alpha_beta_const/2.
        
        A = alpha/(lmax*(lmax+1)) + beta
        B = beta/alpha

        for l in range(0,lmax+1):
            for m in range(0,l+1):
                ## intlm = np.sum( Y[(l,m)] * rho * domg )
                intlm = np.sum( weights * Y[(l,m)] * rho )
                aa[(l,m)] -= A/(1.0 + B*l*(l+1)) * intlm
                
        if self.timers_on: self.timer.stop('UpdateFlowSpectralComponents')
        return aa ## aa[(l,0)]=a0, aa[(l,m>0)].real=ac, aa[(l,m>0)].imag=as 

    def IntegrateOverSurface(self, rh, sqrt_deth, H):
        """
        Integrate the area element, the mean of the expansion and the
        spin components 
        """
        if self.timers_on: self.timer.start('IntegrateOverSurface')
        
        th, ph, weights = self.TPMesh()

        da = weights * sqrt_deth / np.sin(th)
        
        #area  = np.sum( da )
        coarea = np.sum( weights * rh**2 )
        hrms   = np.sum( da * H**2 )
        hmean  = np.sum( da * H )
        #Sx    = np.sum( da * intSx )
        #Sy    = np.sum( da * intSy )
        #Sz    = np.sum( da * intSz )
        
        if self.timers_on: self.timer.stop('IntegrateOverSurface')
        return coarea,hrms,hmean
        
    def MetricDerivatives(self,data,edgeo=1):
        """
        Derivatives of the 3-metric 
        Assume 3D Cartesian coordinates and uniform grid
        """
        if self.timers_on: self.timer.start('MetricDerivatives')
        
        dx = data['coord_x'][1]-data['coord_x'][0]
        dy = data['coord_y'][1]-data['coord_y'][0]
        dz = data['coord_z'][1]-data['coord_z'][0]

        data['gxx_x'],data['gxx_y'],data['gxx_z'] = np.gradient(data['gxx'],dx,dy,dz, edge_order= edgeo)
        data['gxy_x'],data['gxy_y'],data['gxy_z'] = np.gradient(data['gxy'],dx,dy,dz, edge_order= edgeo)
        data['gxz_x'],data['gxz_y'],data['gxz_z'] = np.gradient(data['gxz'],dx,dy,dz, edge_order= edgeo)
        data['gyy_x'],data['gyy_y'],data['gyy_z'] = np.gradient(data['gyy'],dx,dy,dz, edge_order= edgeo)
        data['gyz_x'],data['gyz_y'],data['gyz_z'] = np.gradient(data['gyz'],dx,dy,dz, edge_order= edgeo)
        data['gzz_x'],data['gzz_y'],data['gzz_z'] = np.gradient(data['gzz'],dx,dy,dz, edge_order= edgeo)

        if self.timers_on: self.timer.stop('MetricDerivatives')
        return data

    def DataGridSymm(self,x,y,z, data_sym = None):
        """
        Change interpolation point to account for grid symmetries
        Flag the altered points
        """
        if data_sym == 'bitant':
            symx = np.full(x.shape, False, dtype=bool)
            symy = np.full(y.shape, False, dtype=bool)
            symz = z<0 
            z[z<0] *= - 1.
        elif data_sym == 'octant':
            symx = x<0
            symy = y<0
            symz = z<0
            x[symx] *= - 1.
            y[symy] *= - 1.
            z[symz] *= - 1.
        #elif sym == '':
        #...
        return x,y,z, symx,symy,symz

    def DataCheckLim(self,x,y,z,data):
        """
        Check interpolation point is in the data box
        """
        xmin, xmax = np.amin(data['coord_x']),\
                     np.amax(data['coord_x'])
        ymin, ymax = np.amin(data['coord_y']),\
                     np.amax(data['coord_y'])
        zmin, zmax = np.amin(data['coord_z']),\
                     np.amax(data['coord_z'])
        msg = "Some point outside data box,"
        if x.min() < xmin:
            raise ValueError(msg+" xmin = {}".format(x.min()))
        if x.max() > xmax:
            raise ValueError(msg+" xmax = {}".format(x.max()))
        if y.min() < ymin:
            raise ValueError(msg+" ymin = {}".format(y.min()))
        if y.max() > ymax:
            raise ValueError(msg+" ymax = {}".format(y.max()))
        if z.min() < zmin:
            raise ValueError(msg+" zmin = {}".format(z.min()))
        if z.max() > zmax:
            raise ValueError(msg+" zmax = {}".format(z.max()))
        return

    def _Int3_LinCustom(self, px,py,pz,values, x,y,z):
        """
        Linear 3D interpolation
        (custom cython implementation)
        
        px,py,pz : values of each coordinate (1d arrays)
        values   : function values at px,py,pz (3d array, meshgrid)
        x,y,z    : values at which to interp (3d arrays, mehgrid)
        """
        return Interp3D(values, px,py,pz)(x,y,z)

    def _Int3_LinRGI(self, px,py,pz,values, x,y,z):
        """
        Linear 3D interpolation
        RegularGridInterpolator wrapper
        
        px,py,pz : values of each coordinate (1d arrays)
        values   : function values at px,py,pz (3d array, meshgrid)
        x,y,z    : values at which to interp (3d arrays, mehgrid)
        """
        return RegularGridInterpolator((px,py,pz),values,method='linear')((x,y,z))
    
    # def _Int3_RGI(self, px,py,pz,values, x,y,z, method = 'linear',
    #         processes = 1):
    #     """
    #     RegularGridInterpolator wrapper
    #     Optional: multiprocess pool call
    #     px,py,pz : values of each coordinate (1d arrays)
    #     values   : function values at px,py,pz (3d array, meshgrid)
    #     pti      : values at which to interp [(x0,y0,z0),(x0,y0,z1),...]
    #     """
    #     if processes == 1:
    #         return RegularGridInterpolator((px,py,pz),values,method=method)(pti)
    #     elif processes < 1:
    #         processes = multiprocessing.cpu_count()
    #     rgi = RegularGridInterpolator((px,py,pz),values,method=method)
    #     with multiprocessing.Pool(processes=processes) as pool:
    #         fi = pool.starmap(rgi, pti)
    #     return fi
    
    def MetricInterp(self,center,rh,data,
                     data_sym = None):
        """
        Interpolate 3D Cartesian data to the sphere of radius rh 
        centered in 'center'
        
        Assumes input data

        data['coord_x'] : 1d array with x-coordinates
        data['coord_y'] : 1d array with y-coordinates
        data['coord_z'] : 1d array with z-coordinates
        data['gxx']     : 3d array with gxx component
        data['gxy']     : 3d array with gxy component
        ...
        data['Kxx']     : 3d array with Kxx component
        data['Kxy']     : 3d array with Kxy component
        ...
        data['gxx_x']   : 3d array with x-derivative of gxx 
        ...

        """
        if self.timers_on: self.timer.start('MetricInterp')
        
        th, ph, _ = self.TPMesh()
        sinth = np.sin(th)
        costh = np.cos(th)
        sinph = np.sin(ph)
        cosph = np.cos(ph)
                
        # Global coordinates of the surface
        xc,yc,zc = center
        x = xc + rh *sinth*cosph
        y = yc + rh *sinth*sinph
        z = zc + rh *costh
        
        if data_sym is not None:
            # Correct sphere coords for data grid symmetry
            x,y,z,symx,symy,symz = self.DataGridSymm(x,y,z, data_sym = data_sym)

        # Check sphere coords are in the data grid 
        self.DataCheckLim(x,y,z,data)
        
        # Interp 3D data on the surface
        method = self.interpolation_method
        px,py,pz = data['coord_x'],data['coord_y'],data['coord_z']
        g = {}
        for ij in t_ij:
            g[ij] = self.Int3(px,py,pz, data['g'+ij], x,y,z)
            if not np.isfinite(g[ij]).all():
                raise ValueError("g[{}] not finite".format(ij))
            if data_sym is not None:
                g[ij][symx] = sym2[ij][0] * g[ij][symx]
                g[ij][symy] = sym2[ij][1] * g[ij][symy]
                g[ij][symz] = sym2[ij][2] * g[ij][symz]
        K = {}
        for ij in t_ij:
            K[ij] = self.Int3(px,py,pz, data['K'+ij], x,y,z)
            if not np.isfinite(K[ij]).all():
                raise ValueError("K[{}] not finite".format(ij))
            if data_sym is not None:
                K[ij][symx] = sym2[ij][0] * K[ij][symx]
                K[ij][symy] = sym2[ij][1] * K[ij][symy]
                K[ij][symz] = sym2[ij][2] * K[ij][symz]
            
        dg = {}
        for ij_k in t_ij_k:
            dg[ij_k] = self.Int3(px,py,pz, data['g'+ij_k], x,y,z)
            if not np.isfinite(dg[ij_k]).all():
                raise ValueError("dg[{}] not finite".format(ij_k))
            if data_sym is not None:
                dg[ij_k][symx] = sym3[ij_k][0] * dg[ij_k][symx]
                dg[ij_k][symy] = sym3[ij_k][1] * dg[ij_k][symy]
                dg[ij_k][symz] = sym3[ij_k][2] * dg[ij_k][symz]
        
        if self.timers_on: self.timer.stop('MetricInterp')
        return g, dg, K
    
    def MetricInverse(self,g):
        """
        Inverse 3-metric and determinant
        """
        gxx = g['xx']
        gxy = g['xy']
        gxz = g['xz']
        gyy = g['yy']
        gyz = g['yz']
        gzz = g['zz']
        
        ginvdetxx = gyy*gzz - gyz*gyz
        ginvdetxy = gxz*gyz - gxy*gzz
        ginvdetxz = gxy*gyz - gxz*gyy
        ginvdetyy = gxx*gzz - gxz*gxz
        ginvdetyz = gxy*gxz - gxx*gyz
        ginvdetzz = gxx*gyy - gxy*gxy
        
        detg = gxx*ginvdetxx + gxy*ginvdetxy + gxz*ginvdetxz
        oodetg = 1/detg
        if not np.isfinite(oodetg).all():
            raise ValueError("1/detg not finite")
        
        ginvxx = ginvdetxx *oodetg
        ginvxy = ginvdetxy *oodetg
        ginvxz = ginvdetxz *oodetg
        ginvyy = ginvdetyy *oodetg
        ginvyz = ginvdetyz *oodetg
        ginvzx = ginvdetxz *oodetg
        ginvzz = ginvdetzz *oodetg

        ig = {}
        ig['xx'] = ginvxx
        ig['xy'] = ginvxy
        ig['xz'] = ginvxz
        ig['yy'] = ginvyy
        ig['yz'] = ginvyz
        ig['zz'] = ginvzz

        for ij in t_ij:
            if not np.isfinite(ig[ij]).all():
                raise ValueError("invg[{}] not finite".format(ij))
        
        return detg, ig

    def TraceSym02(self,ig,K):
        """
        Trace of symmetric tensor (0,2)
        """
        ixx = ig['xx']
        ixy = ig['xy']
        ixz = ig['xz']
        iyy = ig['yy']
        iyz = ig['yz']
        izz = ig['zz']

        Kxx = K['xx']
        Kxy = K['xy']
        Kxz = K['xz']
        Kyy = K['yy']
        Kyz = K['yz']
        Kzz = K['zz']
        
        K = ixx*Kxx +  iyy*Kyy +  izz*Kzz\
            + 2.0*(ixy*Kxy + ixz*Kxz +  iyz*Kyz)
        
        return K
        
    def Expansion(self,rh,aa, g,dg,K):
        """
        Compute the expansion function and other quantities

        Metric, metric derivatives and extrinsic curvature are
        expected interpolated on the sphere and in the form

        g['xx'] : 2d array 
        g['xy'] : 2d array
        ...
        dg['xx_x'] : 2d array
        dg['xx_y'] : 2d array
        ...
        K['xx'] : 2d array
        ...

        """
        if self.timers_on: self.timer.start('Expansion')
        
        lmax = self.lmax
        th, ph, _ = self.TPMesh()
        sinth = np.sin(th)
        costh = np.cos(th)
        sinph = np.sin(ph)
        cosph = np.cos(ph)
        
        # Local coordinates of the surface
        xp = rh * sinth * cosph
        yp = rh * sinth * sinph
        zp = rh * costh

        rp   = np.sqrt(xp*xp + yp*yp + zp*zp)
        rhop = np.sqrt(xp*xp + yp*yp)
            
        # First derivatives of (r,theta,phi) with respect to (x,y,z)
        drdx = xp/rp
        drdy = yp/rp
        drdz = zp/rp

        dthetadx = zp*xp/(rp**2*rhop)
        dthetady = zp*yp/(rp**2*rhop)
        dthetadz = -rhop/(rp**2)
            
        dphidx = -yp/(rhop**2)
        dphidy = xp/(rhop**2)
        dphidz = 0.0

        # Second derivatives of (r,theta,phi) with respect to (x,y,z)
        drdxdx = 1.0/rp - xp**2/(rp**3)
        drdxdy = - xp*yp/(rp**3)
        drdxdz = - xp*zp/(rp**3)
        drdydy = 1.0/rp - yp**2/(rp**3)
        drdydz = - yp*zp/(rp**3)
        drdzdz = 1.0/rp - zp**2/(rp**3)
          
        dthetadxdx = zp*(-2.0*xp**4 - xp**2*yp**2 + yp**4+zp**2*yp**2)/(rp**4*rhop**3)
        dthetadxdy = - xp*yp*zp*(3.0*xp**2+3.0*yp**2+zp**2)/(rp**4*rhop**3)
        dthetadxdz = xp*(xp**2+yp**2-zp**2)/(rp**4*rhop)
        dthetadydy = zp*(-2.0*yp**4-yp**2*xp**2+xp**4+zp**2*xp**2)/(rp**4*rhop**3)
        dthetadydz = yp*(xp**2+yp**2-zp**2)/(rp**4*rhop)
        dthetadzdz = 2.0*zp*rhop/(rp**4)
            
        dphidxdx = 2.0*yp*xp/(rhop**4)  
        dphidxdy = (yp**2 - xp**2)/(rhop**4)  
        dphidxdz = 0.0  
        dphidydy = - 2.0*yp*xp/(rhop**4)  
        dphidydz = 0.0  
        dphidzdz = 0.0  

        del rp
        del rhop
        
        # First derivatives of F: dFdi  
        dFdx = drdx
        dFdy = drdy  
        dFdz = drdz

        del drdx
        del drdy
        del drdz
        
        for l in range(0,lmax+1):
            dFdx -= aa[(l,0)] * dthetadx * self.dY_dth[(l,0)]
            dFdy -= aa[(l,0)] * dthetady * self.dY_dth[(l,0)]
            dFdz -= aa[(l,0)] * dthetadz * self.dY_dth[(l,0)]
              
        for l in range(1,lmax+1):
            for m in range(1,l+1):
                dYcdtheta = self.dY_dth[(l,m)].real
                dYsdtheta = self.dY_dth[(l,m)].imag
                dYcdphi = self.dY_dph[(l,m)].real 
                dYsdphi = self.dY_dph[(l,m)].imag 

                ar = aa[(l,m)].real # = ac
                ai = aa[(l,m)].imag # = as

                dFdx -= ar*(dthetadx*dYcdtheta + dphidx*dYcdphi)\
                        + ai*(dthetadx*dYsdtheta + dphidx*dYsdphi)   
                dFdy -= ar*(dthetady*dYcdtheta + dphidy*dYcdphi)\
                        + ai*(dthetady*dYsdtheta + dphidy*dYsdphi)   
                dFdz -= ar*(dthetadz*dYcdtheta + dphidz*dYcdphi)\
                        + ai*(dthetadz*dYsdtheta + dphidz*dYsdphi)

        if not np.isfinite(dFdx).all():
            raise ValueError("dFdx not finite")
        if not np.isfinite(dFdy).all():
            raise ValueError("dFdy not finite")
        if not np.isfinite(dFdz).all():
            raise ValueError("dFdz not finite")
                
        # Second derivatives of F: dFdidj    
        dFdxdx = drdxdx
        dFdxdy = drdxdy  
        dFdxdz = drdxdz  
        dFdydy = drdydy  
        dFdydz = drdydz  
        dFdzdz = drdzdz  

        del drdxdx
        del drdxdy
        del drdxdz
        del drdydy
        del drdydz
        del drdzdz
        
        for l in range(0,lmax+1):
            dY0dtheta = self.dY_dth[(l,0)]
            dY0dthetadtheta = self.d2Y_dth2[(l,0)]
            
            dFdxdx -= aa[(l,0)]*(dthetadxdx*dY0dtheta + dthetadx*dthetadx*dY0dthetadtheta)      
            dFdxdy -= aa[(l,0)]*(dthetadxdy*dY0dtheta + dthetadx*dthetady*dY0dthetadtheta)   
            dFdxdz -= aa[(l,0)]*(dthetadxdz*dY0dtheta + dthetadx*dthetadz*dY0dthetadtheta)   
            #dFdydx -= aa[(l,0)]*(dthetadxdy*dY0dtheta + dthetady*dthetadx*dY0dthetadtheta)   
            dFdydy -= aa[(l,0)]*(dthetadydy*dY0dtheta + dthetady*dthetady*dY0dthetadtheta)   
            dFdydz -= aa[(l,0)]*(dthetadydz*dY0dtheta + dthetady*dthetadz*dY0dthetadtheta)  
            #dFdzdx -= aa[(l,0)]*(dthetadxdz*dY0dtheta + dthetadz*dthetadx*dY0dthetadtheta)   
            #dFdzdy -= aa[(l,0)]*(dthetadydz*dY0dtheta + dthetadz*dthetady*dY0dthetadtheta)    
            dFdzdz -= aa[(l,0)]*(dthetadzdz*dY0dtheta + dthetadz*dthetadz*dY0dthetadtheta)   

        for l in range(1,lmax+1):
            for m in range(1,l+1):
                dYcdtheta = self.dY_dth[(l,m)].real
                dYsdtheta = self.dY_dth[(l,m)].imag
                dYcdphi = self.dY_dph[(l,m)].real 
                dYsdphi = self.dY_dph[(l,m)].imag 
                dYcdthetadtheta = self.d2Y_dth2[(l,m)].real
                dYsdthetadtheta = self.d2Y_dth2[(l,m)].imag
                dYcdphidphi = self.d2Y_dph2[(l,m)].real
                dYsdphidphi = self.d2Y_dph2[(l,m)].imag
                dYcdthetadphi = self.d2Y_dthdph[(l,m)].real 
                dYsdthetadphi = self.d2Y_dthdph[(l,m)].imag 
                
                ar = aa[(l,m)].real # = ac
                ai = aa[(l,m)].imag # = as
                
                #FIXME indentation!
                dFdxdx -= ar*(dthetadxdx*dYcdtheta + dthetadx*(dthetadx*dYcdthetadtheta + dphidx*dYcdthetadphi) + dphidxdx*dYcdphi + dphidx*(dthetadx*dYcdthetadphi + dphidx*dYcdphidphi))\
                        + ai*(dthetadxdx*dYsdtheta + dthetadx*(dthetadx*dYsdthetadtheta + dphidx*dYsdthetadphi) + dphidxdx*dYsdphi + dphidx*(dthetadx*dYsdthetadphi + dphidx*dYsdphidphi))  
                dFdxdy -= ar*(dthetadxdy*dYcdtheta + dthetadx*(dthetady*dYcdthetadtheta + dphidy*dYcdthetadphi) + dphidxdy*dYcdphi + dphidx*(dthetady*dYcdthetadphi + dphidy*dYcdphidphi))\
                        + ai*(dthetadxdy*dYsdtheta + dthetadx*(dthetady*dYsdthetadtheta + dphidy*dYsdthetadphi) + dphidxdy*dYsdphi + dphidx*(dthetady*dYsdthetadphi + dphidy*dYsdphidphi))  
                dFdxdz -= ar*(dthetadxdz*dYcdtheta + dthetadx*(dthetadz*dYcdthetadtheta + dphidz*dYcdthetadphi) + dphidxdz*dYcdphi + dphidx*(dthetadz*dYcdthetadphi + dphidz*dYcdphidphi))\
                        + ai*(dthetadxdz*dYsdtheta + dthetadx*(dthetadz*dYsdthetadtheta + dphidz*dYsdthetadphi) + dphidxdz*dYsdphi + dphidx*(dthetadz*dYsdthetadphi + dphidz*dYsdphidphi))
                #dFdydx -= ar*(dthetadxdy*dYcdtheta + dthetady*(dthetadx*dYcdthetadtheta + dphidx*dYcdthetadphi) + dphidxdy*dYcdphi + dphidy*(dthetadx*dYcdthetadphi + dphidx*dYcdphidphi))\
                #        + ai*(dthetadxdy*dYsdtheta + dthetady*(dthetadx*dYsdthetadtheta + dphidx*dYsdthetadphi) + dphidxdy*dYsdphi + dphidy*(dthetadx*dYsdthetadphi + dphidx*dYsdphidphi))  
                dFdydy -= ar*(dthetadydy*dYcdtheta + dthetady*(dthetady*dYcdthetadtheta + dphidy*dYcdthetadphi) + dphidydy*dYcdphi + dphidy*(dthetady*dYcdthetadphi + dphidy*dYcdphidphi))\
                        + ai*(dthetadydy*dYsdtheta + dthetady*(dthetady*dYsdthetadtheta + dphidy*dYsdthetadphi) + dphidydy*dYsdphi + dphidy*(dthetady*dYsdthetadphi + dphidy*dYsdphidphi))  
                dFdydz -= ar*(dthetadydz*dYcdtheta + dthetady*(dthetadz*dYcdthetadtheta + dphidz*dYcdthetadphi) + dphidydz*dYcdphi + dphidy*(dthetadz*dYcdthetadphi + dphidz*dYcdphidphi))\
                        + ai*(dthetadydz*dYsdtheta + dthetady*(dthetadz*dYsdthetadtheta + dphidz*dYsdthetadphi) + dphidydz*dYsdphi + dphidy*(dthetadz*dYsdthetadphi + dphidz*dYsdphidphi))  
                #dFdzdx -= ar*(dthetadxdz*dYcdtheta + dthetadz*(dthetadx*dYcdthetadtheta + dphidx*dYcdthetadphi) + dphidxdz*dYcdphi + dphidz*(dthetadx*dYcdthetadphi + dphidx*dYcdphidphi))\
                #        + ai*(dthetadxdz*dYsdtheta + dthetadz*(dthetadx*dYsdthetadtheta + dphidx*dYsdthetadphi) + dphidxdz*dYsdphi + dphidz*(dthetadx*dYsdthetadphi + dphidx*dYsdphidphi))
                #dFdzdy -= ar*(dthetadydz*dYcdtheta + dthetadz*(dthetady*dYcdthetadtheta + dphidy*dYcdthetadphi) + dphidydz*dYcdphi + dphidz*(dthetady*dYcdthetadphi + dphidy*dYcdphidphi))\
                #        + ai*(dthetadydz*dYsdtheta + dthetadz*(dthetady*dYsdthetadtheta + dphidy*dYsdthetadphi) + dphidydz*dYsdphi + dphidz*(dthetady*dYsdthetadphi + dphidy*dYsdphidphi))
                dFdzdz -= ar*(dthetadzdz*dYcdtheta + dthetadz*(dthetadz*dYcdthetadtheta + dphidz*dYcdthetadphi) + dphidzdz*dYcdphi + dphidz*(dthetadz*dYcdthetadphi + dphidz*dYcdphidphi))\
                        + ai*(dthetadzdz*dYsdtheta + dthetadz*(dthetadz*dYsdthetadtheta + dphidz*dYsdthetadphi) + dphidzdz*dYsdphi + dphidz*(dthetadz*dYsdthetadphi + dphidz*dYsdphidphi))  

        del dY0dtheta 
        del dY0dthetadtheta

        del dYcdtheta
        del dYsdtheta 
        del dYcdphi 
        del dYsdphi 
        del dYcdthetadtheta 
        del dYsdthetadtheta 
        del dYcdphidphi 
        del dYsdphidphi 
        del dYcdthetadphi
        del dYsdthetadphi
        
        del dthetadx
        del dthetady
        del dthetadz

        del dphidx
        del dphidy
        del dphidz
          
        del dthetadxdx
        del dthetadxdy
        del dthetadxdz
        del dthetadydy
        del dthetadydz
        del dthetadzdz
            
        del dphidxdx
        del dphidxdy
        del dphidxdz
        del dphidydy
        del dphidydz
        del dphidzdz

        if not np.isfinite(dFdxdx).all():
            raise ValueError("dFdxdx not finite")
        if not np.isfinite(dFdxdy).all():
            raise ValueError("dFdxdy not finite")
        if not np.isfinite(dFdxdz).all():
            raise ValueError("dFdxdz not finite")
        if not np.isfinite(dFdydy).all():
            raise ValueError("dFdydy not finite")
        if not np.isfinite(dFdydz).all():
            raise ValueError("dFdydz not finite")
        if not np.isfinite(dFdzdz).all():
            raise ValueError("dFdzdz not finite")
        
        # Metric, K and metric drvts
        detg, ig = self.MetricInverse(g)

        gxx = g['xx']
        gxy = g['xy']
        gxz = g['xz']
        gyy = g['yy']
        gyz = g['yz']
        gzz = g['zz']
        
        ginvxx = ig['xx']
        ginvxy = ig['xy']
        ginvxz = ig['xz']
        ginvyy = ig['yy']
        ginvyz = ig['yz']
        ginvzz = ig['zz']
        
        Kxx = K['xx']
        Kxy = K['xy']
        Kxz = K['xz']
        Kyy = K['yy']
        Kyz = K['yz']
        Kzz = K['zz']

        TrK = self.TraceSym02(ig,K)
        
        dgxxdx = dg['xx_x']
        dgxxdy = dg['xx_y']
        dgxxdz = dg['xx_z']

        dgxydx = dg['xy_x']
        dgxydy = dg['xy_y']
        dgxydz = dg['xy_z']

        dgxzdx = dg['xz_x']
        dgxzdy = dg['xz_y']
        dgxzdz = dg['xz_z']

        dgyydx = dg['yy_x']
        dgyydy = dg['yy_y']
        dgyydz = dg['yy_z']

        dgyzdx = dg['yz_x']
        dgyzdy = dg['yz_y']
        dgyzdz = dg['yz_z']
        
        dgzzdx = dg['zz_x']
        dgzzdy = dg['zz_y']
        dgzzdz = dg['zz_z']

        del g
        del detg        
        del ig
        del K
        del dg
        
        # Expansion
        # ---------
        
        # dFdi with the index up
        dFupdx = ginvxx*dFdx + ginvxy*dFdy + ginvxz*dFdz  
        dFupdy = ginvxy*dFdx + ginvyy*dFdy + ginvyz*dFdz  
        dFupdz = ginvxz*dFdx + ginvyz*dFdy + ginvzz*dFdz  

        # norm of dFdi  
        norm = dFupdx*dFdx + dFupdy*dFdy + dFupdz*dFdz
        
        if not np.isfinite(norm).all():
            raise ValueError("norm not finite")
        
        u = np.zeros_like(norm)
        u[norm>0] = np.sqrt(norm[norm>0])  

        del dFdx
        del dFdy
        del dFdz
        
        # nabla_a nabla_b F  
        nnFxx = dFdxdx - 0.5*(dFupdx*dgxxdx + dFupdy*(2.0*dgxydx-dgxxdy) + dFupdz*(2.0*dgxzdx-dgxxdz))  
        nnFxy = dFdxdy - 0.5*(dFupdx*dgxxdy + dFupdy*dgyydx + dFupdz*(dgyzdx+dgxzdy-dgxydz))  
        nnFxz = dFdxdz - 0.5*(dFupdx*dgxxdz + dFupdy*(dgyzdx + dgxydz - dgxzdy) + dFupdz*dgzzdx)  
        nnFyy = dFdydy - 0.5*(dFupdy*dgyydy + dFupdx*(2.0*dgxydy - dgyydx) + dFupdz*(2.0*dgyzdy-dgyydz))  
        nnFyz = dFdydz - 0.5*(dFupdy*dgyydz + dFupdx*(dgxzdy + dgxydz - dgyzdx) + dFupdz*dgzzdy)  
        nnFzz = dFdzdz - 0.5*(dFupdz*dgzzdz + dFupdx*(2.0*dgxzdz - dgzzdx) + dFupdy*(2.0*dgyzdz-dgzzdy))  

        del dgxxdx
        del dgxxdy
        del dgxxdz

        del dgxydx
        del dgxydy
        del dgxydz

        del dgxzdx
        del dgxzdy
        del dgxzdz

        del dgyydx
        del dgyydy
        del dgyydz

        del dgyzdx
        del dgyzdy
        del dgyzdz
        
        del dgzzdx
        del dgzzdy
        del dgzzdz

        del dFdxdx
        del dFdxdy
        del dFdxdz
        del dFdydy
        del dFdydz
        del dFdzdz
        
        # d2F = g^{ab} nabla_a nabla_b F
        d2F = ginvxx*nnFxx + ginvyy*nnFyy + ginvzz*nnFzz\
              + 2.0*(ginvxy*nnFxy + ginvxz*nnFxz + ginvyz*nnFyz)
        
        del ginvxx
        del ginvxy
        del ginvxz
        del ginvyy
        del ginvyz
        del ginvzz
        
        # dFd^a dFd^b Kab  
        dFdadFdbKab = dFupdx*dFupdx*Kxx + 2.0*dFupdx*dFupdy*Kxy\
                      + 2.0*dFupdx*dFupdz*Kxz + dFupdy*dFupdy*Kyy\
                      + 2.0*dFupdy*dFupdz*Kyz + dFupdz*dFupdz*Kzz
        
        # dFd^a dFd^b nabla_a nabla_b F  
        dFdadFdbFdadb = dFupdx*dFupdx*nnFxx + 2.0*dFupdx*dFupdy*nnFxy\
                        + 2.0*dFupdx*dFupdz*nnFxz + dFupdy*dFupdy*nnFyy\
                        + 2.0*dFupdy*dFupdz*nnFyz + dFupdz*dFupdz*nnFzz  

        # H
        # Eq. 6.7.13 of Alcubierre
        oou = np.zeros_like(u)
        oou[norm>0] = 1.0/u[norm>0]
        
        H = d2F*oou + dFdadFdbKab*oou**2 - dFdadFdbFdadb*oou**3 - TrK  

        del oou
        del TrK
        del d2F
        del dFdadFdbKab
        del dFdadFdbFdadb
        
        # rho = H*u*sigma
        # ---------------
        
        sigma = 1.0  
        rho = H*u*sigma

        # Normal vector
        # -------------
        
        oou = np.zeros_like(u)
        oou[norm>0] = 1.0/u[norm>0]

        Rx = dFupdx *oou
        Ry = dFupdy *oou
        Rz = dFupdz *oou

        if self.timers_on: self.timer.stop('Expansion')
        return H, u, rho, Rx,Ry,Rz

    def SurfaceElement(self,rh,drh_dth,drh_dph,g):
        """
        Determinant of the induced metric on the horizon
        Returns sqrt(h)
        """
        if self.timers_on: self.timer.start('SurfaceElement')
        
        dxdt, dydt, dzdt, dxdp, dydp, dzdp = self.DerivativesXYZvsTP(rh,drh_dth,drh_dph,
                                                                     FDGradient=False)

        gxx = g['xx']
        gxy = g['xy']
        gxz = g['xz']
        gyy = g['yy']
        gyz = g['yz']
        gzz = g['zz']
        
        # Induced metric on the horizon
        h11 = dxdt*(dxdt*gxx+dydt*gxy+dzdt*gxz)\
              + dydt*(dxdt*gxy+dydt*gyy+dzdt*gyz)\
              + dzdt*(dxdt*gxz+dydt*gyz+dzdt*gzz)  
        h12 = dxdt*(dxdp*gxx+dydp*gxy+dzdp*gxz)\
              + dydt*(dxdp*gxy+dydp*gyy+dzdp*gyz)\
              + dzdt*(dxdp*gxz+dydp*gyz+dzdp*gzz)  
        h22 = dxdp*(dxdp*gxx+dydp*gxy+dzdp*gxz)\
              + dydp*(dxdp*gxy+dydp*gyy+dzdp*gyz)\
              + dzdp*(dxdp*gxz+dydp*gyz+dzdp*gzz)  
        
        # Determinant of the induced metric
        sqrt_deth_arg = h11*h22 - h12*h12 
        sqrt_deth_arg[sqrt_deth_arg < 0.0] = 0.0
        if 0 in sqrt_deth_arg:
            raise ValueError("deth <= 0")
        if not np.isfinite(sqrt_deth_arg).all():
            raise ValueError("detg not finite")

        sqrth = np.sqrt(sqrt_deth_arg)

        # Integrate on sphere
        th, _, weights = self.TPMesh()
        da = weights * sqrth / np.sin(th)
        area  = np.sum( da )

        if self.timers_on: self.timer.stop('SurfaceElement')
        return sqrth, area

    def Spin(self,rh,sqrt_deth,u,Rx,Ry,Rz,K):
        """
        Spin integral 
        """
        if self.timers_on: self.timer.start('Spin')
        
        th, ph, weights = self.TPMesh()
        sinth = np.sin(th)
        costh = np.cos(th)
        sinph = np.sin(ph)
        cosph = np.cos(ph)

        # Local coordinates of the surface
        xp = rh * sinth * cosph
        yp = rh * sinth * sinph
        zp = rh * costh
        
        Kxx = K['xx']
        Kxy = K['xy']
        Kxz = K['xz']
        Kyy = K['yy']
        Kyz = K['yz']
        Kzz = K['zz']
                
        # Flat space coordinate rotational killing vectors
        phix_x =  0  
        phix_y = -zp # -(z-zc)  
        phix_z =  yp #  (y-yc)  
        phiy_x =  zp #  (z-zc)  
        phiy_y =  0  
        phiy_z = -xp # -(x-xc)  
        phiz_x = -yp # -(y-yc)  
        phiz_y =  xp #  (x-xc)  
        phiz_z =  0  
        
        # Integrands of S.
        intSx = phix_x*Rx*Kxx + phix_x*Ry*Kxy + phix_x*Rz*Kxz\
                + phix_y*Rx*Kxy + phix_y*Ry*Kyy + phix_y*Rz*Kyz\
                + phix_z*Rx*Kxz + phix_z*Ry*Kyz + phix_z*Rz*Kzz
        intSy = phiy_x*Rx*Kxx + phiy_x*Ry*Kxy + phiy_x*Rz*Kxz\
                + phiy_y*Rx*Kxy + phiy_y*Ry*Kyy + phiy_y*Rz*Kyz\
                + phiy_z*Rx*Kxz + phiy_z*Ry*Kyz + phiy_z*Rz*Kzz
        intSz = phiz_x*Rx*Kxx + phiz_x*Ry*Kxy + phiz_x*Rz*Kxz\
                + phiz_y*Rx*Kxy + phiz_y*Ry*Kyy + phiz_y*Rz*Kyz\
                + phiz_z*Rx*Kxz + phiz_z*Ry*Kyz + phiz_z*Rz*Kzz

        # Integrate
        da = weights * sqrt_deth /( 8.0*np.pi* sinth )
        Sx  = np.sum( da * intSx )
        Sy  = np.sum( da * intSy )
        Sz  = np.sum( da * intSz )

        S = np.sqrt(Sx*Sx + Sy*Sy + Sz*Sz)    

        if self.timers_on: self.timer.stop('Spin')
        return Sx,Sy,Sz,S
    
    def FastFlow(self, time, data,
                 number = 1,
                 data_sym = None):
        """
        Flow loop
        """
        if self.timers_on: self.timer.start('FastFlow')
        
        th, ph, _ = self.TPMesh()
        lmax = self.lmax

        aa = self.InitialGuess(number)
        if self.use_last and self.last_rh[number]>0.:
            aa[(0,0)] = self.last_rh[number]
        meanradius = aa[(0,0)]/np.sqrt(4.0*np.pi)
        cpoint = self.CentralPoint(number)

        mass = 0.0
        
        if self.verbose:
            print("\nSearching for horizon {} at time {}".format(number,time))
            print("centered in (x,y,z) = {}\nr_mean = {}".format(cpoint,meanradius))
            print("---"*30)
            print(" iter   area   mass   meanradius  hmean")
            print("---"*30)
            
        # ------------------------------------------------------------
        for it in range(self.flow_iter):
            if it == self.flow_iter-1:
                self.ahfound[number] = False
                print(" Failed: {} = {}".format(it,self.flow_iter))
                break

            rh,drh_dth,drh_dph = self.RadiusFromSphericalHarmonics(aa)

            g,dg,K = self.MetricInterp(cpoint,rh,data,
                                       data_sym = data_sym)

            H,u,rho,Rx,Ry,Rz = self.Expansion(rh,aa,g,dg,K)

            sqrth,area = self.SurfaceElement(rh,drh_dth,drh_dph,g)

            Sx,Sy,Sz,S = self.Spin(rh,sqrth,u,Rx,Ry,Rz,K)
            
            coarea,hrms,hmean = self.IntegrateOverSurface(rh,sqrth,H)
            
            mass_old = mass
            mass = np.sqrt(area/(16.0*np.pi))
            meanradius = aa[(0,0)]/np.sqrt(4.0*np.pi)
            hrms = hrms/area
            
            if self.verbose:
                print(" {:04d} {:.12e} {:.12e} {:.12e} {:.12e}"
                      .format(it,area,mass, meanradius, hmean))

            if np.fabs(hmean) > self.hmean_tol:
                self.ahfound[number] = False
                print(" Failed: hmean > {}".format(self.hmean_tol))
                break
        
            if meanradius < 0.0:
                self.ahfound[number] = False
                print(" Failed: surface negative height")
                break

            # End flow when mass difference is small
            if np.fabs(mass_old-mass) < self.mass_tol :
                self.ahfound[number] = True
                break
        
            # Find new spectral components
            aa = self.UpdateFlowSpectralComponents(aa,rho, 
                                                   alpha_beta_const = self.alpha_beta_const)
        # ------------------------------------------------------------

        if self.ahfound[number]:
            self.last_rh[number] = aa[(0,0)]
            if self.verbose:
                print("Horizon {} found:".format(number))
                print(" mass   = {:.12e}".format(mass))
                print(" rmean  = {:.12e}".format(meanradius))
                print(" hrms   = {:.12e}".format(hrms))
                print(" hmean  = {:.12e}".format(hmean))
                print(" Sx     = {:.12e}".format(Sx))
                print(" Sy     = {:.12e}".format(Sy))
                print(" Sz     = {:.12e}".format(Sz))
                print(" S      = {:.12e}".format(S))
                print(" E      = {:.12e}".format(np.sqrt(mass**2 + 0.25*S**2/mass**2)))
                print(" a      = {:.12e}".format(S/mass**2))
            if self.output:
                self.OutputResults(number,time,cpoint,mass,Sx,Sy,Sz,S,coarea)
            if self.output_modes:
                self.OutputModes(number,aa)
            if self.output_iterations:
                self.OutputIterations(number,time,it)
            if self.output_surface:
                self.OutputSurface(number,aa,time)
            
        if self.timers_on: self.timer.stop('FastFlow')
        return

    def OutputModes(self,number,aa):
        """
        Write modes
        Overwrite the previous file!
        """
        fname = "{}/horizon{:02d}_modes.txt".format(self.outputdir,number)
        with open(fname, 'w') as f:
            f.write("{:d}".format(self.lmax))
            for l in range(0,self.lmax+1):
                for m in range(0,l+1):
                    f.write("{:12e} ".format(aa[(l,m)]))
        return
    
    def OutputResults(self,number,
                      time,cpoint,mass,Sx,Sy,Sz,S,coarea):
        """
        Write result
        Append if file exists
        """
        xc,yc,zc = cpoint
        data = np.c_[time,xc,yc,zc,mass,Sx,Sy,Sz,S,coarea]
        fname = "{}/horizon{:02d}.txt".format(self.outputdir,number)
        if not os.path.isfile(fname):
            header="time,xc,yc,zc,mass,Sx,Sy,Sz,S,coarea"
            np.savetxt(fname, data, header = header, fmt='%.12e')
        else:
            with open(fname, "ab") as f:
                np.savetxt(f, data, fmt='%.12e')
        return
    
    def OutputIterations(self,number,time,niter):
        """
        Write iterations 
        Append if file exists
        """
        data = np.c_[time,niter]
        fname = "{}/horizon{:02d}_iters.txt".format(self.outputdir,number)
        if not os.path.isfile(fname):
            header="time,iterations"
            np.savetxt(fname, data, header = header)
        else:
            with open(fname, "ab") as f:
                np.savetxt(f, data)
        return    

    def OutputSurface(self,number,aa,time):
        """
        Write surface file
        """
        rh,_,_ = self.RadiusFromSphericalHarmonics(aa)
        th, ph, _ = self.TPMesh()
        sinth = np.sin(th)
        costh = np.cos(th)
        sinph = np.sin(ph)
        cosph = np.cos(ph)
                
        # Global coordinates of the surface
        xc,yc,zc = self.CentralPoint(number)
        x = xc + rh *sinth*cosph
        y = yc + rh *sinth*sinph
        z = zc + rh *costh

        size = np.size(x)
        shape = np.shape(x)
        data = np.c_[x.flatten(),y.flatten(),z.flatten()]
        fname = "{}/horizon{:02d}_surface.txt".format(self.outputdir,number)
        header="time = {:.8e}\npoints = {} x {} = {}\n 1:x 2:y 3:z".format(time,shape[0],shape[1],size)
        np.savetxt(fname, data, header = header)
        return
    
    def InputData(self,fname,
                  input_data_from = 'HDF5_simple', data_sym = None,
                  options = None):
        """
        Convert input HDF5 datasets into the data structure used in
        MetricInterp()  

        data['time'] : time of the snapshot
        data['coord_x'] : 1d array with x-coordinates
        data['coord_y'] : 1d array with y-coordinates
        data['coord_z'] : 1d array with z-coordinates
        data['gxx']     : 3d array with gxx component
        data['gxy']     : 3d array with gxy component
        ...
        data['Kxx']     : 3d array with Kxx component
        data['Kxy']     : 3d array with Kxy component
        ...
        data['gxx_x']   : (optional) 3d array with x-derivative of gxx 
        data['gxx_y']   : (optional) 3d array with y-derivative of gxx 
        ...

        See data.py for possible methods.
        """
        if input_data_from == 'data':
            # fname is directly the data
            return fname
        elif input_data_from == 'HDF5_simple':
            return read_h5_dset(fname)
        elif input_data_from == 'HDF5_athena':
            return read_h5_dset_athena(fname, data_sym = data_sym, **options)
        # ... add here ...
        else:
            raise ValueError("unknown {}".format(method))
        
    def run(self, filenames,
            input_data_from = 'HDF5_simple',
            data_sym = None,
            data_options = None):
        """
        Search the AHF in hypersurfaces from a list of data files, 
        one file per time slice

        Input data can be passed with different methods.
        """
        if self.timers_on: self.timer.start('run')
        
        for fin in filenames:

            # get data and check they are good
            data = self.InputData(fin,
                                  input_data_from = input_data_from,
                                  data_sym = data_sym,
                                  options = data_options)
            for k in data.keys():
                if not np.isfinite(data[k]).all():
                    raise ValueError("{} in data is not finite".format(k))

            time = data['time']
            if time < self.tbeg and time > self.tend:
                continue
            if self.compute_metric_drvts:
                data = self.MetricDerivatives(data)
            for number in range(self.nhorizons):
                # Search for this horizon 
                self.FastFlow(time, data,
                              number = number,
                              data_sym = data_sym)

        if self.timers_on:
            self.timer.stop('run')
            self.timer.summary(ToScreen = self.verbose,
                               ToFile="timers.txt")
        return
    

if __name__ == "__main__":
    
    
    p = Puncture()
    p.Output("puncture.h5")

    ahf = AHF(initial_radius = [1.5],
              outputdir='test')
    ahf.run(['puncture.h5'])
    
