import os.path, sys
import numpy as np
import h5py

if 'AHFpy' in sys.modules:
    from AHFpy.data import ab_indexes, ij_indexes
    from AHFpy.data import write_h5_dset
else:
    # Assume we are in ahfpy/AHFpy/
    from data import ab_indexes, ij_indexes
    from data import write_h5_dset

import matplotlib.pyplot as plt

class Kerr():
    """
    Kerr data in Cartesian coordinates
    """
    
    def __init__(self,
                 M = 1., # Mass parameter
                 a = 0., # Spin parameter
                 x = np.arange(-5+0.125/2,5,0.25/2), # better staggered!
                 y = np.arange(-5+0.125/2,5,0.25/2),
                 z = np.arange(-5+0.125/2,5,0.25/2) ):

        self.M = M
        self.a = a
        
        self.x = x
        self.y = y
        self.z = z
        
        self.grid = self.CartGrid()
        
    def CartGrid(self):
        """
        Cartesian mesh grid
        """
        return np.meshgrid(self.x,self.y,self.z,
                           indexing="ij")
    
    def SphereGrid(self):
        """
        Spherical Polar mesh grid
        """
        x,y,z = self.grid 
        rho = np.sqrt(x**2 + y**2)
        r = np.sqrt(x**2 + y**2 + z**2)
        th = np.arctan2(rho,z)
        ph = np.arctan2(y,x)
        return r,th,ph

    def FlatMetric(self):
        """
        Minkowski metric (-,+,+,+)
        """
        g = {}
        for (a,b) in ab_indexes:
            g[(a,b)] = +1 if a == b else 0.
        g[(0,0)] = -1
        return g

    def Coefficients(self):
        """
        Eq(62-67)
        https://arxiv.org/pdf/gr-qc/9707050
        """
        rb,th,ph = self.SphereGrid()        
        costh2, sinth2 = np.cos(th)**2, np.sin(th)**2
        div_rb2 = 1.0/rb**2
        div_rb4 = 1.0/rb**4
        
        M,a = self.M, self.a
        a2  = a**2
        rh = 0.5*np.sqrt(self.M**2-self.a**2)

        r = M + rb + rh**2/rb
        r2 = r**2
        Delta = r2 - 2*M*r + a2
        rho2 = r2 + a2 * costh2
        rho = np.sqrt(r2 + a2 * costh2)
        rho3 = rho2*rho

        div_fact = 1.0/np.sqrt((r2+a2)**2 - Delta*a2*sinth2)
        
        D = div_fact * 2*a**3 *M*r*np.sqrt(Delta)*np.cos(th) * div_rb4/(rho3)
        C = div_fact * a*M*( 2*r2*(r2+a2) + rho2*(r2-a2) ) * div_rb4/(rho3)
        B = (rho2 + 2*M*r) * a2 * div_rb4/rho2
        A = rho2 * div_rb2
        
        return A,B,C,D
  
    def Vectors_vw(self,n_u=(0,0,0,1)):
        """
        Eq(63)
        https://arxiv.org/pdf/gr-qc/9707050
        """
        x,y,z=self.grid
        rb = np.sqrt(x**2+y**2+z**2)
        X = (0,x,y,z)
        costh = 0.0
        for i in range(1,4):
            costh += n_u[i] * X[i]
        costh = costh/rb
        _,_,C,D = self.Coefficients()
        v_u = {}
        v_u[0] = 0.0
        v_u[1] = n_u[2]*X[3] - n_u[3]*X[2] 
        v_u[2] = n_u[3]*X[1] - n_u[1]*X[3] 
        v_u[3] = n_u[1]*X[2] - n_u[2]*X[1] 
        w_u = {}
        for i in range(1,4):
            w_u[i] = C * X[i] + D * (n_u[i] - costh * X[i])
        return v_u, w_u 
    
    def Metric(self):
        """
        Metric $g_{ij}$
        """
        eta = self.FlatMetric()
        A,B,_,_ = self.Coefficients()
        v_d, _ = self.Vectors_vw()
        g = {}
        for (i,j) in ij_indexes:
            g[(i,j)] = A*( eta[(i,j)] + B*v_d[i]*v_d[j] )
        return g

    def ExtrinsicCurvature(self):
        """
        Extrinsic cuvrvature $K_{ij}$
        """
        v_d, w_d = self.Vectors_vw()
        K = {}
        for (i,j) in ij_indexes:
            K[(i,j)] = v_d[i]*w_d[j] + w_d[i]*v_d[j]
        return K
        
    def Output(self,fname):
        """
        Write HDF5 data
        """
        points = self.x, self.y, self.z
        g = self.Metric()
        K = self.ExtrinsicCurvature()
        return write_h5_dset(fname,
                             time = 0, points = points,
                             g = g, K = K)
    
    
if __name__ == "__main__":


    # a=0 M=1, full grid
    # ------------------------------
    k = Kerr()
    k.Output("kerr_a00.h5")
    
    # a=0.9 M=1, full grid
    # ---------------------------------
    k = Kerr(a=0.9)
    k.Output("kerr_a09.h5")

        

        
