import os.path, sys
import numpy as np
import h5py

if 'AHFpy' in sys.modules:
    from AHFpy.data import ab_indexes, ij_indexes
    from AHFpy.data import write_h5_dset
else:
    # Assume we are in ahfpy/AHFpy/
    from data import ab_indexes, ij_indexes
    from data import write_h5_dset

#TODO check Kij expression from Alcubierre
#TODO check consistency of Kij analytical - FD 
#TODO check consistency of transformed Kij(sph) components with cartesian (a=0)
#TODO cleanup

import matplotlib.pyplot as plt

class KerrSchild():
    """
    Kerr-Schild data in Cartesian coordinates
    """
    
    def __init__(self,
                 M = 1., # Mass parameter
                 a = 0., # Spin parameter
                 x = np.arange(-5+0.125/2,5,0.25/2), # better staggered!
                 y = np.arange(-5+0.125/2,5,0.25/2),
                 z = np.arange(-5+0.125/2,5,0.25/2) ):

        self.M = M
        self.a = a
        
        self.x = x
        self.y = y
        self.z = z
        
        self.grid = self.CartGrid()
        
    def CartGrid(self):
        """
        Cartesian mesh grid
        """
        return np.meshgrid(self.x,self.y,self.z,
                           indexing="ij")
    
    def SphereGrid(self):
        """
        Spherical Polar mesh grid
        """
        x,y,z = self.grid 
        rho = np.sqrt(x**2 + y**2)
        r = np.sqrt(x**2 + y**2 + z**2)
        th = np.arctan2(rho,z)
        ph = np.arctan2(y,x)
        return r,th,ph

    def FlatMetric(self):
        """
        Minkowski metric (-,+,+,+)
        """
        g = {}
        for (a,b) in ab_indexes:
            g[(a,b)] = +1 if a == b else 0.
        g[(0,0)] = -1
        return g

    def Rfun(self):
        """
        KS Radial function
        """
        x,y,z = self.grid 
        rho2 = x**2 + y**2 + z**2
        a2  = self.a**2
        R2 = 0.5*( rho2 - a2 + np.sqrt((rho2-a2)**2+4*a2*z**2) )
        #R2 = 0.5 * (rho2 - a2) + np.sqrt(0.25 * (rho2 - a2)**2 + a2 * z**2)
        return np.sqrt(R2), R2

    def Hfun(self):
        """
        KS H function 
        """
        x,y,z = self.grid 
        R,_ = self.Rfun()
        return self.M * R**3/(R**4 + self.a**2 * z**2) 
  
    def NullCovl_d(self):
        """
        KS Null covector $l_a$
        """
        x,y,z = self.grid 
        a = self.a
        R,R2 = self.Rfun()
        divR = 1.0/(R2 + a**2)
        l_d = {}
        l_d[0] = np.ones_like(x)
        l_d[1] = (R*x + a*y)*divR
        l_d[2] = (R*y - a*x)*divR
        l_d[3] = z/R
        return l_d
    
    def Metric(self):
        """
        Metric $g_{ab} = \eta_{ab} + 2 H l_a l_b$
        """
        eta = self.FlatMetric()
        l_d = self.NullCovl_d()
        H = self.Hfun()
        g = {}
        for (a,b) in ab_indexes:
            g[(a,b)] = eta[(a,b)] + 2.0*H*l_d[a]*l_d[b]
        return g

    def ExtrinsicCurvature(self):
        """
        Extrinsic cuvrvature $K_{ij}$
        Eq.(3.4.37) of Alcubierre's book
        """
        if self.a != 0.0:
            raise ValueError("Kij implemented only for a=0")
        M = self.M
        R,_ = self.Rfun()
        eta = self.FlatMetric() 
        l_d = self.NullCovl_d()
        H = self.Hfun()
        alpha = np.sqrt(1.0/(1.0 + 2.0*H)) 
        K = {}
        for (i,j) in ij_indexes:
            K[(i,j)] = 2.0*H*alpha/R*(eta[(i,j)] - (2.0 + H) * l_d[i]*l_d[j])
        return K

    def Lapse(self):
        """
        Lapse function $\alpha$
        """
        H = self.Hfun()
        return np.sqrt(1.0/(1.0 + 2.0*H)) 

    def Shift(self):
        """
        Shift function $\beta^i$ #TODO
        """
        eta = self.FlatMetric() 
        l_d = self.NullCovl_d()
        #l_u = {}
        H = self.Hfun()
        beta_u = {}
        #for k in range(1,4):
        #    beta_u[k] = -2.0*H*l_u[0]*l_u[k]/() 
        return None
    
    def Metric3from4(self,g, test=True):
        """
        Build lapse, shift and 3-metric $\gamma_{ij}$ from 4-metric $g_{ab}$
        """
        if test: print('Test Metric3from4:')
        gamma_dd = {}
        for (i,j) in ij_indexes:
            gamma_dd[(i,j)] = g[(i,j)]
        gamma_uu,_ = self.InvMetric3(gamma_dd,test=test)
        beta_d = {}
        beta_d[0] = 0.0
        for i in range(1,4):
            beta_d[i] = g[(0,i)]
        beta_u = {}
        beta_u[0] = 0.0
        for i in range(1,4):
            beta_u[i] = 0.0
            for j in range(1,4):
                ij_sym = tuple(sorted((i,j)))
                beta_u[i] += gamma_uu[ij_sym] * beta_d[j]
        beta2 = 0.0
        for i in range(1,4):
            beta2 += beta_u[i]*beta_d[i]
        alpha = np.sqrt(beta2 - g[(0,0)])
        if test:
            print('3+1 components:')
            print('g_00 finite  ? {}'.format(np.isfinite(g[(0,0)]).all()))
            #print('g_00 < 0     ? {}'.format((g[(0,0)]<0).all())) 
            print('alpha finite ? {}'.format(np.isfinite(alpha).all()))
            for i in range(1,4):
                print('beta^{} finite? {}'\
                      .format(i,np.isfinite(beta_u[i]).all()))
            for (i,j) in ij_indexes:
                print('gamma_{}{} finite? {}'\
                      .format(i,j,np.isfinite(gamma_dd[(i,j)]).all()))
        return alpha,beta_u,gamma_dd

    def DetMetric3(self,gamma_dd, test=True):
        """
        Determinant of 3-metric $\gamma_{ij}$
        """
        detgam = - gamma_dd[(1,3)]**2 * gamma_dd[(2,2)] \
            + 2.0*gamma_dd[(1,2)] * gamma_dd[(1,3)] * gamma_dd[(2,3)] \
            - gamma_dd[(1,1)] * gamma_dd[(2,3)]**2 \
            - gamma_dd[(1,2)]**2 * gamma_dd[(3,3)] \
            + gamma_dd[(1,1)] * gamma_dd[(2,2)] * gamma_dd[(3,3)]
        if test:
            print('Test DetMetric3:')
            print('detgam finite? {}'.format(np.isfinite(detgam).all()))
            print('detgam > 0   ? {}'.format((detgam>0).all()))
        return detgam

    def InvMetric3(self,gamma_dd, detgam=None, test=True):
        """
        Inverse 3-metric $\gamma^{ij}$
        """
        gamma_uu = {}
        gamma_uu[(1,1)] = gamma_dd[(2,2)]*gamma_dd[(3,3)]\
            - gamma_dd[(2,3)]*gamma_dd[(2,3)];
        gamma_uu[(1,2)] = gamma_dd[(1,3)]*gamma_dd[(2,3)]\
            - gamma_dd[(1,2)]*gamma_dd[(3,3)];
        gamma_uu[(1,3)] = gamma_dd[(1,2)]*gamma_dd[(2,3)]\
            - gamma_dd[(1,3)]*gamma_dd[(2,2)];
        gamma_uu[(2,2)] = gamma_dd[(1,1)]*gamma_dd[(3,3)]\
            - gamma_dd[(1,3)]*gamma_dd[(1,3)];
        gamma_uu[(2,3)] = gamma_dd[(1,2)]*gamma_dd[(1,3)]\
            - gamma_dd[(1,1)]*gamma_dd[(2,3)];
        gamma_uu[(3,3)] = gamma_dd[(1,1)]*gamma_dd[(2,2)]\
            - gamma_dd[(1,2)]*gamma_dd[(1,2)];

        if detgam is None:
            detgam =\
                  gamma_dd[(1,1)]*gamma_uu[(1,1)]\
                + gamma_dd[(1,2)]*gamma_uu[(1,2)]\
                + gamma_dd[(1,3)]*gamma_uu[(1,3)]
            oodetgam = 1.0/detgam
            for (i,j) in ij_indexes:
                gamma_uu[(i,j)] *= oodetgam
                
        if test:
            print('Test InvMetric3:')
            print('detgam finite? {}'.format(np.isfinite(detgam).all()))
            print('detgam > 0   ? {}'.format((detgam>0).all()), end=' ')
            _detgam = self.DetMetric3(gamma_dd,test=False)
            maxaDD = np.max(np.absolute(_detgam - detgam))
            print('  Methods compare ? {}  (max_diff = {:.2e})'\
                  .format(np.allclose(_detgam,detgam),maxaDD))
            for (i,j) in ij_indexes:
                print('gamma^{}{} finite? {}'\
                      .format(i,j,np.isfinite(gamma_uu[(i,j)]).all()))
            I = {}
            for (i,j) in ij_indexes:
                I [(i,j)] = 0.0
                for k in range(1,4): 
                    ik_sym = tuple(sorted((i,k)))
                    jk_sym = tuple(sorted((j,k)))
                    I[(i,j)] += gamma_uu[ik_sym] * gamma_dd[jk_sym]
            eta = self.FlatMetric()
            for (i,j) in ij_indexes:
                print('I^{}_{} = gamma^{}k gamma_{}k ? {}'\
                      .format(i,j,i,j,np.allclose(I[(i,j)],eta[(i,j)])))           
        return gamma_uu, detgam

    def Metric4from3(self,alpha,beta_u,gamma_dd):
        """
        Build 4-metric $g_{ab}$ from 3-metric $\gamma_{ij}$, 
        lapse $\alpha$ and shift $\beta^i$
        """
        beta_d = {}
        beta_d[0] = 0.0
        for i in range(1,4):
            beta_d[i] = 0.0
            for j in range(1,4):
                ij_sym = tuple(sorted((i,j)))
                beta_d[i] += gamma_dd[ij_sym]*beta_u[j]
        g = {}
        for (i,j) in ij_indexes:
            g[(i,j)] = gamma_dd[(i,j)]
        g[(0,0)] = - alpha**2
        for i in range(1,4):
            g[(0,0)] += beta_d[i] * beta_u[i]
            g[(0,i)] = beta_d[i]
        return g
    
    def ExtrinsicCurvature_FD(self,alpha,beta_u,gamma_dd, test=True):
        """
        Calculate $K_{ij}$ numerically from definition
        ($\dot{\gamma}_{ij}=0$)
        """
        x,y,z = self.x,self.y,self.z
        dx,dy,dz = np.diff(x)[0], np.diff(y)[0], np.diff(z)[0]
        K = {}
        for (i,j) in ij_indexes:

            dgamij_d = {}
            dgamij_d[1], dgamij_d[2], dgamij_d[3] =\
                np.gradient(gamma_dd[(i,j)], dx,dy,dz, edge_order = 2)
            beta_delgamm = 0.0
            for k in range(1,4):
                beta_delgamm += beta_u[k] * dgamij_d[k]

            gamm_delbeta = 0.0
            for k in range(1,4):
                ik_sym = tuple(sorted((i,k)))
                kj_sym = tuple(sorted((k,j)))
                dbetak_d = {}
                dbetak_d[1], dbetak_d[2], dbetak_d[3] =\
                    np.gradient(beta_u[k], dx,dy,dz, edge_order = 2)
                gamm_delbeta +=\
                      gamma_dd[ik_sym] * dbetak_d[i]\
                    + gamma_dd[kj_sym] * dbetak_d[j]
            
            K[(i,j)] = (beta_delgamm + gamm_delbeta)/(2.0*alpha)
        #
        if test and self.a==0.0:
            atol = np.amax([dx,dy,dz]) # this is really an upper bound! FD are O(dx^2)
            print("Test Kij from FD with Kij analytical expression (atol={}):"\
                  .format(atol))
            _K = self.ExtrinsicCurvature()
            for ij in ij_indexes:
                print('{} Finite? {}  '.format(ij, np.isfinite(K[ij]).all()),end=' ')
                maxaKK = np.max(np.absolute(_K[ij] - K[ij]))
                print('Close? {}   max_diff = {:.2e}'\
                      .format(np.allclose(_K[ij],K[ij],rtol=0,atol=atol),maxaKK))
        # 
        return K
        
    def MetricSphericalPolar(self):
        """
        3+1 Metric $g_{ab}$ in spherical polar coordinates
        Appendix A of https://arxiv.org/abs/1704.00599

        NB The paper has a typo in the definition of rho coordinate (in the text):
           rho(paper) = r2 + a2 * costh2
        it should instead read:
            rho^2 := r2 + a2 * costh2
        This is usually called Sigma in BL coordinates,
        and it is the gamma_thth metric coef, see Eq. (A.3).
        The code is actually correct:
        https://bitbucket.org/zach_etienne/wvuthorns/src/7802e89105d8b077e848e62a2e10ed50a79ed8cd/ShiftedKerrSchild/src/ShiftedKerrSchild.c#lines-134
        """
        M = self.M
        a = self.a 
        a2 = a**2
        r,th,ph = self.SphereGrid() # (r,theta,phi) from self.grid
        r2, costh2, sinth2 = r**2, np.cos(th)**2, np.sin(th)**2
        rho2 = r2 + a2 * costh2 # This is correct.
        twoMr_rho2 = 2*M*r/rho2
        alpha = 1.0/(np.sqrt(1.0 + twoMr_rho2))# (A.1)
        alpha2 = alpha**2
        beta_u = {}
        beta_u[0] = 0
        beta_u[1] = alpha2 * twoMr_rho2        # (A.2)
        beta_u[2] = 0                          # (A.3)
        beta_u[3] = 0                          # (A.3)
        gamma_dd = {}
        gamma_dd[(1,1)] = 1.0 + twoMr_rho2        # (A.4)
        gamma_dd[(1,2)] = 0                       # (A.3)
        gamma_dd[(1,3)] = -a*gamma_dd[(1,1)]*sinth2  # (A.5)
        gamma_dd[(2,2)] = rho2                    # (A.3)
        gamma_dd[(2,3)] = 0                       # (A.7)
        gamma_dd[(3,3)] = (r2 + a2 + twoMr_rho2*a2*sinth2) * sinth2 # (A.7)
        return alpha, beta_u, gamma_dd 

    def ExtrinsicCurvatureSphericalPolar(self, test=False):
        """
        Extrinsic cuvrvature $K_{ij}$ in spherical polar coordinates
        Appendix A of https://arxiv.org/abs/1704.00599
        https://bitbucket.org/zach_etienne/wvuthorns/src/7802e89105d8b077e848e62a2e10ed50a79ed8cd/ShiftedKerrSchild/src/ShiftedKerrSchild.c#lines-266
        """
        M = self.M
        a = self.a 
        a2 = a**2
        a4 = a**4
        r,th,ph = self.SphereGrid()
        costh, sinth = np.cos(th), np.sin(th)
        costwoth = np.cos(2*th)
        r2, costh2, sinth2 = r**2, costh**2, sinth**2
        rho2 = r2 + a2 * costh2 # This is correct.
        # Appendix A https://arxiv.org/abs/1704.00599
        twoMr = 2*M*r
        twoMr_rho2 = twoMr/rho2
        A = a2*(costwoth + 1.0) + 2*r2             # (A.8)
        B = A + 2*twoMr                            # (A.9)
        D = np.sqrt(twoMr/(a2*costh2 + r2) + 1.0)  # (A.10)        
        K = {}
        for (a,b) in ab_indexes:
            K[(a,b)] = 0.0
        K[(1,1)] = (D*(A+twoMr))/(A**2*B) * (4*M*(a2*costwoth+a2-2*r2)) # (A.11)
        K[(1,2)] = D/(A*B) * (4*a2*twoMr*sinth*costh)                   # (A.12)
        K[(1,3)] = D/A**2  * (-2*a*M*sinth2*(a2*costwoth+a2-2*r2))      # (A.13)
        K[(2,2)] = D/B     * 4*M*r2                                     # (A.14)
        K[(2,3)] = D/(A*B) * (-8*a**3*M*r*sinth**3*costh)               # (A.15)
        K[(3,3)] = D/(A**2*B) * (\
                                 twoMr*sinth2*( a4*(r-M)*np.cos(4*th) + a4*(M+3*r)\
                                                + 4*a2*r2*(2*r-M)
                                                + 4*a2*r *costwoth*(a2+r*(M+2*r))
                                                + 8*r**5\
                                               )\
                                 )                                      # (A.16-17)
        # sym (not nedded)
        #K[(2,1)] = K[(1,2)]
        #K[(3,1)] = K[(1,3)]
        #K[(3,2)] = K[(2,3)]
        if test:
            # TEST: The following is taken from the public code:
            mcrfact1 = np.sqrt(1 + (2*M*r)/rho2)
            arcfact1 = (a2 + 2*r*(2*M + r) + a2*costwoth)
            arcfact2 = (a2 + 2*r2 + a2*costwoth)**2
            #
            print('Test two expressions for Kij:')
            _K = {}
            for (a,b) in ab_indexes:
                _K[(a,b)] = 0.0      
            _K[(1,1)]= (-4*M*mcrfact1*(-a2 + 2*r2 - a2*costwoth)*(a2 + 2*r*(M + r) + a2*costwoth))/(arcfact2*arcfact1);
            _K[(1,2)]= (4*a2*M*r*mcrfact1*np.sin(2*th))/((a2 + 2*r2 + a2*costwoth)*arcfact1);
            _K[(1,3)]= (-2*a*M*mcrfact1*(a2 - 2*r2 + a2*costwoth)*sinth2)/arcfact2;
            _K[(2,2)]= (4*M*r2*mcrfact1)/arcfact1;
            _K[(2,3)]= (-8*a**3*M*r*costh*mcrfact1*sinth**3)/((a2 + 2*r2 + a2*costwoth)*arcfact1);
            _K[(3,3)]= (2*M*r*mcrfact1*(8*r**5 + 4*a2*r2*(-M + 2*r) + a4*(M + 3*r) + 4*a2*r*(a2 + r*(M + 2*r))*costwoth + a4*(-M + r)*np.cos(4*th))*sinth2)/(arcfact2*arcfact1);
            for ij in ij_indexes:
                print('{} : {}'.format(ij,np.allclose(_K[ij],K[ij])))
        return K

    def Transform3TensorComponents(self,T,
                                   TransType='Sph2Car',
                                   TensorType='(02)'):
        """
        Transform 3-tensors components from/to spherical polar to/from Cartesian 
        Support for TensorType 
         = 10   vector $V^i$
         = (02) symmetric 02-tensor $g_{ij} = g_{(ij)}$

        https://en.wikipedia.org/wiki/List_of_common_coordinate_transformations#From_spherical_coordinates
        """
        if TensorType not in ['10','(02)']:
            raise ValueError("transformation of tensor {} not implemented".format(TensorType))
        if TransType not in ['Sph2Car','Car2Sph']:
            raise ValueError("transformation {} not implemented".format(TransType))
    
        x,y,z = self.grid 
        rho2 = x**2 + y**2
        rho  = np.sqrt(rho2)
        r2   = x**2 + y**2 + z**2
        r    = np.sqrt(r2)

        _,th,ph = self.SphereGrid() # (theta,phi) from self.grid
        costh, sinth = np.cos(th), np.sin(th)
        cosph, sinph = np.cos(ph), np.sin(ph)
        
        if TransType == 'Sph2Car':
            J = {}
            J[(0,0)] = 1.0
            J[(0,1)] = J[(1,0)] = 0.0
            J[(0,2)] = J[(2,0)] = 0.0
            J[(0,3)] = J[(3,0)] = 0.0
            J[(1,1)] = x/r
            J[(1,2)] = y/r
            J[(1,3)] = z/r
            J[(2,1)] = x*z/(r2*rho)
            J[(2,2)] = y*z/(r2*rho)
            J[(2,3)] = -rho/r2
            J[(3,1)] = -y/(rho2)
            J[(3,2)] =  x/(rho2)
            J[(3,3)] = 0.0

        if TransType == 'Car2Sph':
            J = {}
            J[(0,0)] = 1.0
            J[(0,1)] = J[(1,0)] = 0.0
            J[(0,2)] = J[(2,0)] = 0.0
            J[(0,3)] = J[(3,0)] = 0.0
            J[(1,1)] =  sinth*cosph
            J[(1,2)] =  r*costh*cosph
            J[(1,3)] = -r*sinth*sinph
            J[(2,1)] =  sinth*sinph
            J[(2,2)] =  r*costh*sinph
            J[(2,3)] =  r*sinth*cosph
            J[(3,1)] =  costh
            J[(3,2)] = -r*sinth
            J[(3,3)] =  0.0
        
        # New components
        if TensorType == '10':
            Tp = {}
            Tp[0] = 0.0
            for ip in range(1,4):
                Tp[ip] = 0.0
                for i in range(1,4):
                    Tp[ip] += J[(i,ip)] * T[i]
        
        if TensorType == '(02)':
            # sort tuple to account for symmetry
            Tp = {}
            for (ip,jp) in ij_indexes:
                Tp[(ip,jp)] = 0.0
                for i in range(1,4):
                    for j in range(1,4):
                        ij_sym = tuple(sorted((i,j)))
                        Tp[(ip,jp)] += J[(i,ip)] * J[(j,jp)] * T[ij_sym]
        return Tp
    
    def Calculate(self,comp_from='Cartesian'):
        """
        Calculate the $g_{ab}$ and $K_{ab}$ in Cartesian components
        from either Cartesian or spherical polar expressions
        """
        if comp_from == 'SphericalPolar':
            alpha, beta_u, gamma_dd = self.MetricSphericalPolar()
            gamma_dd = self.Transform3TensorComponents(gamma_dd,
                                                       TransType='Sph2Car',
                                                       TensorType='(02)')
            beta_u = self.Transform3TensorComponents(beta_u,
                                                     TransType='Sph2Car',
                                                     TensorType='10')
            g = self.Metric4from3(alpha,beta_u,gamma_dd) #Ok
            K_dd = self.ExtrinsicCurvatureSphericalPolar()
            K = self.Transform3TensorComponents(K_dd,
                                                TransType='Sph2Car',
                                                TensorType='(02)')
 
            # #### DEBUG
            """
            alpha, beta_u, gamma_dd = self.Metric3from4(g) 
            K_FD = self.ExtrinsicCurvature_FD(alpha,beta_u,gamma_dd)
            print('Test SphericalPolar Kij calculation against FD (a=0):')
            for c in K.keys():
                maxaKK = np.max(np.absolute(K_FD[c] - K[c]))
                print(' {} : {} max_diff = {}'.format(c,np.allclose(K[c],K_FD[c]),maxaKK))
            """
            # #####

        else:
            g = self.Metric() 
            K = self.ExtrinsicCurvature()
            alpha, beta_u, gamma_dd = self.Metric3from4(g) #det and inv are Ok, Ok?
            K_FD = self.ExtrinsicCurvature_FD(alpha,beta_u,gamma_dd) #TODO make this default, with auto check against analytical formula for a==0

            # #### DEBUG
            x,y,z = self.x,self.y,self.z
            dx,dy,dz = np.diff(x)[0], np.diff(y)[0], np.diff(z)[0]
            iz = np.shape(alpha)[2]//2
            iy = np.shape(alpha)[1]//2
            for k in range(1,4):
                dbetak_d = {}
                dbetak_d[1], dbetak_d[2], dbetak_d[3] =\
                    np.gradient(beta_u[k], dx,dy,dz, edge_order = 2)
                fig, ax = plt.subplots()
                ax.semilogy(x,np.absolute(dbetak_d[1])[:,iy,iz])
                ax.semilogy(x,np.absolute(dbetak_d[2])[:,iy,iz])
                ax.semilogy(x,np.absolute(dbetak_d[3])[:,iy,iz])

            for ij in ij_indexes:
                #fig, ax = plt.subplots()
                #cf = ax.contourf(np.absolute(K[ij]-K_FD[ij])[:,:,iz])
                #cbar = fig.colorbar(cf)
                fig, ax = plt.subplots()
                ax.semilogy(x,np.absolute(K[ij])[:,iy,iz],'k')
                ax.semilogy(x,np.absolute(K_FD[ij])[:,iy,iz])
                ax.semilogy(x,np.absolute(K[ij]-K_FD[ij])[:,iy,iz],'--')
            plt.show()
            # ####
            
        return g,K
    
    def Output(self,fname,comp_from='Cartesian'):
        """
        Write HDF5 data
        """
        points = self.x, self.y, self.z
        g,K = self.Calculate(comp_from=comp_from)
        return write_h5_dset(fname,
                             time = 0, points = points,
                             g = g, K = K)
    
    
if __name__ == "__main__":


    do_plots = False

    
    # KS data for a=0 M=1, full grid
    # ------------------------------
    dfile = "kerrschild.h5"
    ks = KerrSchild()
    ks.Output(dfile)

    g = ks.Metric()
    K = ks.ExtrinsicCurvature()
    xyz = ks.grid
    alpha = ks.Lapse()

    if do_plots:
        iz = np.shape(alpha)[2]//2
        iy = np.shape(alpha)[1]//2
        #print(np.min(alpha[:,:,iz]))
    
        fig, ax = plt.subplots()
        ax.contourf(xyz[0][:,:,iz], xyz[1][:,:,iz], g[(1,1)][:,:,iz])
        ax.set(title=r"gxx a={} M={} (Cart)".format(ks.a,ks.M))
        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.grid()
        fig.savefig("kerrschild_gxx.png")
                
        fig, ax = plt.subplots()
        cf_ = ax.contourf(xyz[0][:,:,iz], xyz[1][:,:,iz], alpha[:,:,iz])
        cbar = fig.colorbar(cf_)
        ax.set(title=r"$\alpha$ $a={}$ $M={}$ (Cart)".format(ks.a,ks.M))
        ax.set_xlabel('x')
        ax.set_ylabel('z')
        ax.grid()
        fig.savefig("kerrschild_alpha.png")
                
        fig, ax = plt.subplots()
        cf_ = ax.contourf(xyz[0][:,:,iz], xyz[1][:,:,iz], K[(1,1)][:,:,iz])
        cbar = fig.colorbar(cf_)
        ax.set(title=r"Kxx $a={}$ $M={}$ (Cart)".format(ks.a,ks.M))
        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.grid()
        fig.savefig("kerrschild_Kxx.png")

        fig, ax = plt.subplots()
        cf_ = ax.contourf(xyz[0][:,iy,:], xyz[2][:,iy,:], K[(1,1)][:,iy,:])
        cbar = fig.colorbar(cf_)
        ax.set(title=r"Kxx $a={}$ $M={}$ (Cart)".format(ks.a,ks.M))
        ax.set_xlabel('x')
        ax.set_ylabel('z')
        ax.grid()
        fig.savefig("kerrschild_Kxx.png")
        
        plt.show()                 

    # test read data
    # --------------
    from data import read_h5_dset
    data = read_h5_dset(dfile)
    print("Test HDF5 reading\n arrays_equal: {}".format(np.array_equal(g[(1,1)],data['gxx'])))


    # KS data with bitant symmetry
    # ----------------------------
    #ksb = KerrSchild(x = np.arange(-5+0.125,5,0.25), 
    #                 y = np.arange(-5+0.125,5,0.25),
    #                 z = np.arange(-0.125,5,0.25))
    #ksb.Output("kerrschild_bitant.h5")

    
    # KS data with octant symmetry
    # ----------------------------
    """
    kso = KerrSchild(x = np.arange(-0.125/2,5,0.25/2), 
                     y = np.arange(-0.125/2,5,0.25/2),
                     z = np.arange(-0.125/2,5,0.25/2))
    kso.Output("kerrschild_octant.h5")

    if do_plots:
        g = kso.Metric()
        K = kso.ExtrinsicCurvature()
        b_d = kso.Shift_d()
        xyz = kso.grid
        alpha = kso.Lapse()

        iz = 1
        print(np.min( alpha[:,:,iz] ))

        fig, ax = plt.subplots()
        contourf_ = ax.contourf(xyz[0][:,:,iz], xyz[1][:,:,iz], alpha[:,:,iz])
        cbar = fig.colorbar(contourf_)
        ax.set(title=r"$\alpha$ a={} M={} (Cart)".format(kso.a,kso.M))
        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.grid()
        fig.savefig("kerrschild_alpha.png")
        plt.show()                 
    """

    # test data from spherical polar computation
    # ------------------------------------------
    """
    dfile = "kerrschild_sp.h5"    
    ks_sp = KerrSchild()
    ks_sp.Output(dfile, comp_from="SphericalPolar")
    spdata = read_h5_dset(dfile)

    print('Test SphericalPolar calculation against Cartesian (a=0):')
    for c in data.keys():
        print(' {} : {}'.format(c,np.allclose(data[c],spdata[c])))
                                
    if do_plots:
        iz = np.shape(spdata['gxx'])[0]//2
        fig, ax = plt.subplots()
        cf_ = ax.contourf(spdata['coord_x'], spdata['coord_y'], spdata['gxx'][:,:,iz])
        cbar = fig.colorbar(cf_)
        ax.set(title=r"gxx a={} M={} (Sph)".format(ks_sp.a,ks_sp.M))
        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.grid()
        plt.show()

        fig, ax = plt.subplots()
        cf_ = ax.contourf(spdata['coord_x'], spdata['coord_y'], spdata['Kxx'][:,:,iz])
        ax.set_xlabel('x')
        ax.set_ylabel('y')
        cbar = fig.colorbar(cf_)
        ax.set(title=r"Kxx $a={}$ $M={}$ (Sph)".format(ks_sp.a,ks_sp.M))
        ax.grid()
        plt.show()
    """
    
    # KS data for a=0.9 M=1, full grid
    # ---------------------------------
    #ks = KerrSchild(a=0.9)
    #ks.Output("kerrschild_sp_09.h5", comp_from="SphericalPolar")


    # KS data for a=0.9 M=1, bitant grid
    # ----------------------------------
    #ks = KerrSchild(a=0.9,
    #                x = np.arange(-5+0.125,5,0.25), 
    #                y = np.arange(-5+0.125,5,0.25),
    #                z = np.arange(-0.125,5,0.25))
    #ks.Output("kerrschild_sp_09_bitant.h5", comp_from="SphericalPolar")

        

        
