import os.path, sys
import numpy as np
import h5py

if 'AHFpy' in sys.modules:
    from AHFpy.data import ab_indexes, ij_indexes
    from AHFpy.data import write_h5_dset
else:
    # Assume we are in ahfpy/AHFpy/
    from data import ab_indexes, ij_indexes
    from data import write_h5_dset

import matplotlib.pyplot as plt

class Puncture():
    """
    Puncture data in Cartesian coordinates
    """
    
    def __init__(self,
                 M = 1., # Mass parameter
                 x = np.arange(-5+0.125/2,5,0.25/2), # better staggered!
                 y = np.arange(-5+0.125/2,5,0.25/2),
                 z = np.arange(-5+0.125/2,5,0.25/2)):

        self.M = M

        self.x = x
        self.y = y
        self.z = z
        
        self.grid = None        
        self.grid = self.CartGrid()
        
    def CartGrid(self):
        """
        Cartesian mesh grid
        """
        #if not self.grid == None:
        #    return self.grid
        return np.array(np.meshgrid(self.x,self.y,self.z,
                           indexing="ij"))
        
    def Psi4(self):
        """
        Conformal factor
        """
        x,y,z = self.CartGrid()
        r = np.sqrt(x**2 + y**2 + z**2)
        return (1+0.5*self.M/r)**4

    def Lapse(self,precollapsed=False):
        """
        Lapse function
        """
        if precollapsed:
            return 1.0/np.sqrt(self.Psi4())
        else:
            M = self.M
            x,y,z = self.CartGrid()
            r = np.sqrt(x**2 + y**2 + z**2)
            return (1-M/(2*r))/(1+M/(2*r))

    def FlatMetric(self):
        """
        Minkowski metric (-,+,+,+)
        """
        g = {}
        for (a,b) in ab_indexes:
            g[(a,b)] = +1 if a == b else 0.
        g[(0,0)] = -1
        return g
    
    def Metric(self):
        """
        Metric $g_{ab}$
        """
        eta = self.FlatMetric()
        Psi4 = self.Psi4()
        g = {}
        for (a,b) in ab_indexes:
            g[(a,b)] = Psi4 * eta[(a,b)]
        g[(0,0)] = - self.Lapse()**2
        return g

    def ExtrinsicCurvature(self):
        """
        Extrinsic cuvrvature $K_{ij}$
        """
        x,y,z = self.CartGrid()
        K = {}
        for (i,j) in ij_indexes:
            K[(i,j)] = 0*x
        return K
        
    def Output(self,fname):
        """
        Write HDF5 data
        """
        g = self.Metric()
        K = self.ExtrinsicCurvature()
        points = self.x, self.y, self.z
        return write_h5_dset(fname,
                             time = 0, points = points,
                             g = g, K = K)
    
    
if __name__ == "__main__":


    masses = [0.25,1.0]
    
    for M in masses:
        h = M/8.
        xx = np.arange(-4*M+h*M/2,4*M,h*M)
        p = Puncture(M=M, x = xx, y = xx, z = xx)
        p.Output("puncture_m{0:03}.h5".format(int(100*M)))

    g = p.Metric()
    xyz = p.grid
    alpha = p.Lapse()
    
    iz = np.shape(alpha)[0]//2
    print(np.min( alpha[:,:,iz] ))
    
    fig, ax = plt.subplots()
    ax.contourf(xyz[0][:,:,iz], xyz[1][:,:,iz], g[(1,1)][:,:,iz])
    ax.set(title=r"$g_{xx}$")
    ax.grid()
    fig.savefig("puncture_gxx.png")

    fig, ax = plt.subplots()
    contourf_ = ax.contourf(xyz[0][:,:,iz], xyz[1][:,:,iz], alpha[:,:,iz])
    cbar = fig.colorbar(contourf_)
    ax.set(title=r"$\alpha$")
    ax.grid()
    fig.savefig("puncture_alpha.png")
    plt.show()
