import sys, os
import h5py
import numpy as np


""" Tensors 
"""

# indexes for symmetric 3-tensors
ij_indexes = [(1,1),(1,2),(1,3),
              (2,2),(2,3),
              (3,3)]

# indexes for symmetric 4-tensors
ab_indexes = [(0,0),(0,1),(0,2),(0,3),
              (1,1),(1,2),(1,3),
              (2,2),(2,3),
              (3,3)]

# labels for tensor components variables
t_ij = ['xx','xy','xz','yy','yz','zz']
t_ij_k = ['xx_x','xx_y','xx_z',
          'xy_x','xy_y','xy_z',
          'xz_x','xz_y','xz_z',
          'yy_x','yy_y','yy_z',
          'yz_x','yz_y','yz_z',
          'zz_x','zz_y','zz_z']

# symmetries for symmetric 3-tensors
# sign under reflection in directions
# x,y,z = 0,1,2
# See: BAM main/main/tensors.c
sym2 = {}
sym2['xx'] = [+1,+1,+1]
sym2['xy'] = [-1,-1,+1]
sym2['xz'] = [-1,+1,-1]
sym2['yy'] = [+1,+1,+1]
sym2['yz'] = [+1,-1,-1]
sym2['zz'] = [+1,+1,+1]

# symmetries for symmetric 3-tensors derivatives 
sym3 = {}
sym3['xx_x'] = [-1,+1,+1] 
sym3['xy_x'] = [+1,-1,+1] 
sym3['xz_x'] = [+1,+1,-1] 
sym3['yy_x'] = [-1,+1,+1] 
sym3['yz_x'] = [-1,-1,-1] 
sym3['zz_x'] = [-1,+1,+1] 
sym3['xx_y'] = [+1,-1,+1] 
sym3['xy_y'] = [-1,+1,+1] 
sym3['xz_y'] = [-1,-1,-1] 
sym3['yy_y'] = [+1,-1,+1] 
sym3['yz_y'] = [+1,+1,-1] 
sym3['zz_y'] = [+1,-1,+1] 
sym3['xx_z'] = [+1,+1,-1] 
sym3['xy_z'] = [-1,-1,-1] 
sym3['xz_z'] = [-1,+1,+1] 
sym3['yy_z'] = [+1,+1,-1] 
sym3['yz_z'] = [+1,-1,+1] 
sym3['zz_z'] = [+1,+1,-1]


""" Routines to import data
"""

def write_h5_dset(fname, time = None, points = None,
                  g = None, dg = None, K = None):
    """
    Helper to write HDF5 (example)

    Can be used to pass data as 'HDF5_simple'
    """
    fields = {'coord': ['time','coord_x','coord_y','coord_z'],
              'metric': ['gxx','gxy','gxz','gyy','gyz','gzz'],
              'curvature': ['Kxx','Kxy','Kxz','Kyy','Kyz','Kzz'],
              'metric_drvts': ['gxx_x','gxx_y','gxx_z',
                               'gxy_x','gxy_y','gxy_z',
                               'gxz_x','gxz_y','gxz_z',
                               'gyy_x','gyy_y','gyy_z',
                               'gyz_x','gyz_y','gyz_z',
                               'gzz_x','gzz_y','gzz_z']}
    groups = list(fields.keys())
    with h5py.File(fname, 'w') as fn:

        if points is not None:
            grp = groups[0]
            fn.create_group(grp)
            fn[grp].create_dataset(name='coord_x', data=points[0])
            fn[grp].create_dataset(name='coord_y', data=points[1])
            fn[grp].create_dataset(name='coord_z', data=points[2])
            if time is not None:
                fn[grp].create_dataset(name='time', data=time)

        if g:
            grp = groups[1]
            fn.create_group(grp)
            ds = iter(fields[grp])
            for (i,j) in ij_indexes:
                fn[grp].create_dataset(name=next(ds), data=g[(i,j)])
        if K:
            grp = groups[2]
            fn.create_group(grp)
            ds = iter(fields[grp])
            for (i,j) in ij_indexes:
                fn[grp].create_dataset(name=next(ds), data=K[(i,j)])

        if dg:
            grp = groups[3]
            fn.create_group(grp)
            ds = iter(fields[grp])
            for (i,j) in ij_indexes:
                for k in range(1,4):
                    fn[grp].create_dataset(name=next(ds), data=g[(i,j,k)])

    return


def read_h5_dset(fname):
    """
    Helper to read HDF5 (example)

    Convert HDF5 datasets to the following data structure:

    data['time'] : time of the snapshot
    data['coord_x'] : 1d array with x-coordinates
    data['coord_y'] : 1d array with y-coordinates
    data['coord_z'] : 1d array with z-coordinates
    data['gxx']     : 3d array with gxx component
    data['gxy']     : 3d array with gxy component
    ...
    data['Kxx']     : 3d array with Kxx component
    data['Kxy']     : 3d array with Kxy component
    ...
    data['gxx_x']   : (optional) 3d array with x-derivative of gxx 
    data['gxx_y']   : (optional) 3d array with y-derivative of gxx 
    ...

    Can be used to pass data as 'HDF5_simple'
    """
    data = {}
    fields = ['time','coord_x','coord_y','coord_z',
              'gxx','gxy','gxz','gyy','gyz','gzz',
              'Kxx','Kxy','Kxz','Kyy','Kyz','Kzz',
              'gxx_x','gxx_y','gxx_z',
              'gxy_x','gxy_y','gxy_z',
              'gxz_x','gxz_y','gxz_z',
              'gyy_x','gyy_y','gyy_z',
              'gyz_x','gyz_y','gyz_z',
              'gzz_x','gzz_y','gzz_z']
    # Look for the above fields as datasets of any group
    with h5py.File(fname,'r') as fn:
        for group in fn.keys():
            for field in fn[group].keys():
                if field in fields:
                    data[field] = fn[group][field][()]
    return data


def read_h5_dset_athena(fname, data_sym = None,
                        num_ghost = -1, level = None, initial_radius = None):
    """
    Helper to read HDF5 from Athena++

    Convert HDF5 datasets to the following data structure:

    data['time'] : time of the snapshot
    data['coord_x'] : 1d array with x-coordinates
    data['coord_y'] : 1d array with y-coordinates
    data['coord_z'] : 1d array with z-coordinates
    data['gxx']     : 3d array with gxx component
    data['gxy']     : 3d array with gxy component
    ...
    data['Kxx']     : 3d array with Kxx component
    data['Kxy']     : 3d array with Kxy component
    ...
    data['gxx_x']   : (optional) 3d array with x-derivative of gxx 
    data['gxx_y']   : (optional) 3d array with y-derivative of gxx 
    ...

    Can be used to pass data as 'HDF5_athena'
    """
    from AHFpy.data_athena import athdf, octant_to_full

    data = {}
    fields = ['gxx','gxy','gxz','gyy','gyz','gzz',
              'Kxx','Kxy','Kxz','Kyy','Kyz','Kzz']
    if initial_radius == None:
        data_ath = athdf(fname, num_ghost = num_ghost, vertex_centered = True, dtype = np.float64, level = level)
    else:
        data_ath = athdf(fname, num_ghost = num_ghost, vertex_centered = True, dtype = np.float64, level = level, x1_max = initial_radius, x2_max = initial_radius, x3_max = initial_radius, x1_min = -initial_radius, x2_min = -initial_radius, x3_min = -initial_radius)
    data['time'] = data_ath["Time"]
    data['coord_x'] = data_ath["x1f"][0:-1]
    data['coord_y'] = data_ath["x2f"][0:-1]
    data['coord_z'] = data_ath["x3f"][0:-1]
    
    for fi in fields:
        ## Swap axes because Athena output is zyx. 'C' order after that for Cython
        data[fi] = np.swapaxes(data_ath["adm.{}".format(fi)], 0, 2).copy(order='C')
    return data



""" 3D linear interpolation in cython
"""

if 'AHFpy' in sys.modules:
    from AHFpy.interp_3d import interp3D
else:
    # Assume we are in ahfpy/AHFpy/
    import pyximport
    pyximport.install(setup_args={"script_args" : ["--verbose"],
                                  "include_dirs":np.get_include()})
    from interp_3d import interp3D

class Interp3D(object):
    def __init__(self, v, x, y, z):
        self.v = v
        self.min_x, self.max_x = x[0], x[-1]
        self.min_y, self.max_y = y[0], y[-1]
        self.min_z, self.max_z = z[0], z[-1]
        self.delta_x = (self.max_x - self.min_x)/(x.shape[0]-1)
        self.delta_y = (self.max_y - self.min_y)/(y.shape[0]-1)
        self.delta_z = (self.max_z - self.min_z)/(z.shape[0]-1)
    def __call__(self, xi, yi, zi):
        # xi,yi,zi are the meshgrid output
        xs = (xi.flatten()-self.min_x)/self.delta_x
        ys = (yi.flatten()-self.min_y)/self.delta_y
        zs = (zi.flatten()-self.min_z)/self.delta_z   

        return interp3D(self.v, xs, ys, zs).reshape(xi.shape)

    
""" Utilities
"""

import time

class Timers():
    """
    Provides simple timers
    """
    def __init__(self, outputdir = '.'):

        self.outputdir = outputdir
        self.timer = [] # list (dict) of timers

    def find(self,name):
        """
        Return timer and index
        """
        idx,t = next(((i,item) for (i,item) in enumerate(self.timer)\
                      if item["name"] == name), (-1,False))
        return idx,t

    def start(self,name):
        """
        Start a timer, create entry if not exists
        """
        idx,t = self.find(name)
        if t:
            t['start'] = time.time()
            t['calls'] += 1
            self.timer[idx].update(t)
        else:
            t = {'name': name, 
                 'start': time.time(),
                 'delta': 0.,
                 'total': 0.,
                 'calls': 1}
            self.timer.append(t)
        return
        
    def stop(self,name):
        """
        Stop a timer
        """
        idx,t = self.find(name)
        if t:
            t['delta'] = time.time() - t['start']
            t['total'] += t['delta']
            self.timer[idx].update(t)
        else:
            raise ValueError("Timer '{}' does not exists".format(name))
        return
    
    def show(self,name):
        """
        Print timer info now
        """
        idx,t = self.find(name)
        if t:
            for k, v in t.items():
                if k == 'name': continue
                print("Timer({}) {} : {}".format(name,k, v))
        else:
            raise ValueError("Timer '{}' does not exists".format(name))
        return

    def summary(self,ToScreen=True,ToFile=False):
        """
        Print summary info 
        """
        t_sorted = sorted(self.timer, key=lambda k: k['total']) 
        s = "# Timing info\n"
        s += "# Name Calls TotTime Time/Call\n"
        for t in t_sorted:
            s += "{:<30} {:09d} {:.9e} {:.9e}\n"\
                 .format(t['name'],t['calls'],t['total'],t['total']/t['calls'])
        if ToFile:
            with open(self.outputdir+'/'+ToFile,'a') as f:
                f.write(s)
        if ToScreen:
            print(s)
        return



