"""
Adapted from: 
https://stackoverflow.com/questions/41220617/python-3d-interpolation-speedup
https://github.com/jglaser/interp3d

python setup.py build_ext --inplace
"""

import numpy as np
cimport numpy as np
from libc.math cimport floor
from cython cimport boundscheck, wraparound, nonecheck, cdivision
from cython.parallel import prange

@boundscheck(False)
@wraparound(False)
@nonecheck(False)
def interp3D(np.float_t[:,:,::1] v, np.float_t[::1] xs, np.float_t[::1] ys, np.float_t[::1] zs):

    cdef int Nx, Ny, Nz, dim
    Nx, Ny, Nz = v.shape[0], v.shape[1], v.shape[2]
    dim = xs.size
    cdef np.ndarray[np.float_t, ndim=1] interpolated = np.zeros(dim, dtype=np.float)

    _interp3D(&v[0,0,0], &xs[0], &ys[0], &zs[0], &interpolated[0], Nx, Ny, Nz, dim)
    return interpolated

@cdivision(True)
cdef inline void _interp3D(np.float_t *v, np.float_t *x_points, np.float_t *y_points, np.float_t *z_points, np.float_t *result, int Nx, int Ny, int Nz, int dim):

    cdef:
        int i, x0, x1, y0, y1, z0, z1, Nyz
        np.float_t x, y, z, xd, yd, zd, c00, c01, c10, c11, c0, c1, c

    Nyz = Ny*Nz

    #for i in range(dim):
    for i in prange(dim, nogil=True):
        x = x_points[i]
        y = y_points[i]
        z = z_points[i]

        x0 = <int>floor(x)
        x1 = x0 + 1
        y0 = <int>floor(y)
        y1 = y0 + 1
        z0 = <int>floor(z)
        z1 = z0 + 1

        xd = (x-x0)/(x1-x0)
        yd = (y-y0)/(y1-y0)
        zd = (z-z0)/(z1-z0)

        if x0 >= 0 and y0 >= 0 and z0 >= 0:
            c00 = v[Nyz*x0+Nz*y0+z0]*(1-xd) + v[Nyz*x1+Nz*y0+z0]*xd
            c01 = v[Nyz*x0+Nz*y0+z1]*(1-xd) + v[Nyz*x1+Nz*y0+z1]*xd
            c10 = v[Nyz*x0+Nz*y1+z0]*(1-xd) + v[Nyz*x1+Nz*y1+z0]*xd
            c11 = v[Nyz*x0+Nz*y1+z1]*(1-xd) + v[Nyz*x1+Nz*y1+z1]*xd

            c0 = c00*(1-yd) + c10*yd
            c1 = c01*(1-yd) + c11*yd

            c = c0*(1-zd) + c1*zd

        else:
            c = 0

        result[i] = c 
