# Test AHF with Kerr-Schild data

import sys, os
import numpy as np
import matplotlib.pyplot as plt

if 'AHFpy' in sys.modules:
    from AHFpy import AHF
    from AHFpy.KerrSchild import KerrSchild
else:
    sys.path.insert(0, os.path.abspath('../AHFpy'))
    from AHF import AHF
    from KerrSchild import KerrSchild

# Generate KS black-hole data
# ---------------------------

dfile = "data/kerrschild.h5"

ks = KerrSchild(M=1., a=0.)
ks.Output(dfile)

# Search for the horizon
# ----------------------

outputdir = os.path.basename(sys.argv[0]).split('.')[0]

ahf = AHF(flow_iter = 100, # Max iterations for the fast flow
          nphi = 8, # Number of points in phi direction
          ntheta = 4, # Number of circles in theta direction
          lmax = 4, # Maximum number of Legendre polynomials
          mass_tol = 1e-3, # Mass tolerance, stopping criterion for flow
          hmean_tol = 100.,  # h-mean tolerance
          time = (0, 1), # Time window to search
          nhorizons = 1, # Such many surfaces (MOTS) will be searched 
          use_last = True, # use the last surface found as initial guess
          initial_guess_expand = 1.01, # factor to make initial guess larger
          initial_radius = [2.5], # list of guesses, one per horizon
          central_points = [(0.,0.,0.)], # list of central points
          interpolation_method = 'linear', # order of interpolation onto surface"  
          compute_metric_drvts = True, # Compute first derivative of the 3-metric
          output = True, # output results about MOTS
          output_modes = True, # output spherical modes of MOTS
          outputdir = outputdir,                 
          verbose = True,
          timers_on = True)

ahf.run([dfile])

# Check coefficients
# ------------------

clm = np.loadtxt(os.path.join(outputdir,'horizon00_modes.txt'),
                 dtype = np.complex128)

xlm = []
for l in np.arange(0,ahf.lmax+1):
    for m in np.arange(0,l+1):
        xlm.append('{}{}'.format(str(l),str(m)))

fig, ax = plt.subplots()
ax.semilogy(xlm,np.abs(clm.real),'o', markerfacecolor="None", label='Real')
ax.semilogy(xlm,np.abs(clm.imag),'x', label='Imag')
ax.set(xlabel=r'$(\ell,m)$', ylabel=r'$|c_{\ell m}|$')
ax.grid()
ax.legend()
fig.savefig(os.path.join(outputdir,'coefficients.png'))
plt.show()

