# Test 3d interpolation with tri-linear function

import sys, os
import numpy as np
import scipy as sp
import itertools
import time

def func(x,y,z):
    """
    Function to interpolate
    Exact for first order polys
    """
    return (x-1)*(y+2)*(z-3)

def error(xig,yig,zig,fi,interp=None):
    """
    Calculate error
    """
    fg = func(xig,yig,zig)
    er = fg - fi
    rtol, atol = 1e-5, 1e-8
    print("Error {}:".format(interp))
    if np.allclose(er,0., rtol=rtol, atol=atol):
        print(" all-close to rtol = {} atol = {}".format(rtol,atol))
    else:
        print(" abs.err.  rel.err. true_val.  interp_val.")
        for (i,e) in enumerate(er.flatten()):
            fi_i, fg_i = fi.flatten()[i],fg.flatten()[i]
            print("{:+.2e} {:+.1e} {:+.3e} {:+.3e}".format(e,e/fg_i,fg_i,fi_i))
    return

# data
# ----

x = np.arange(-1,1,0.1)
y = np.arange(-1,1,0.1)
z = np.arange(-1,1,0.1)

xg,yg,zg = np.meshgrid(x,y,z,
                       indexing='ij')

points = list(itertools.product(x,y,z))
values = func(xg,yg,zg) 

print("Data shape / size: {} / {}".format(xg.shape,xg.size))

# new grid points
# ---------------

# single point tests
#xig,yig,zig = np.meshgrid(0.5,0.4,0.3) # these corresponds to data pts
#xig,yig,zig = np.meshgrid(0.512345676,0.423423456,0.32545566) 

# meshgrid test
# xi = np.arange(0.1,0.9,0.212)
# yi = np.arange(0.1,0.9,0.313)
# zi = np.arange(0.1,0.9,0.2567)
# xig,yig,zig = np.meshgrid(xi,yi,zi, indexing='ij') 

# sphere test
r = 0.6
ntheta = 4
nphi = 3
dt = np.pi/ntheta
dp = 2.0*np.pi/nphi
th = dt *(0.5 + np.arange(ntheta))
ph = dp *(0.5 + np.arange(nphi))
th,ph = np.meshgrid(th,ph)

xig = r * np.sin(th) * np.cos(ph)
yig = r * np.sin(th) * np.sin(ph)
zig = r * np.cos(th)

xyzi = list(zip(xig.flatten(),yig.flatten(),zig.flatten()))

#for x_i,y_i,z_i in xyzi:
#    print(x_i,y_i,z_i)
print("New points shape / size : {} / {}".format(xig.shape,xig.size))

# interp: griddata
# ----------------
from scipy.interpolate import griddata

Int = "griddata"

tic = time.perf_counter()
fi = griddata(points, values.flatten(), (xig,yig,zig), method='linear')
toc = time.perf_counter()

print('------------------------------')
print("Timing {}: {:0.6f} secs".format(Int,toc-tic))
error(xig,yig,zig,fi,interp=Int)

# interp: RegularGridInterpolator
# -------------------------------
from scipy.interpolate import RegularGridInterpolator

Int = "RegularGridInterpolator"

tic = time.perf_counter()
fi = RegularGridInterpolator((x,y,z), values, method='linear')(xyzi)
toc = time.perf_counter()

print('------------------------------')
print("Timing {}: {:0.6f} secs".format(Int,toc-tic))
fi = np.reshape(fi,xig.shape)
error(xig,yig,zig,fi,interp=Int)

# interp: custom cython
# ---------------------
if 'AHFpy' in sys.modules:
    from AHFpy.data import Interp3D
else:
    sys.path.insert(0, os.path.abspath('../AHFpy'))
    from data import Interp3D

Int = "CythonInterp3D"

tic = time.perf_counter()
f = Interp3D(values, x,y,z)
fi = f(xig,yig,zig)
toc = time.perf_counter()

print('------------------------------')
print("Timing {}: {:0.6f} secs".format(Int,toc-tic))
error(xig,yig,zig,fi,interp=Int)

