# Test 3d interpolation using Kerr-Schild data

import numpy as np

# Generate this file by running AHFpy/KerrSchild.py
dfile = 'kerrschild.h5'


# read data
# ---------
from AHFpy.data import read_h5_dset
data = read_h5_dset(dfile)


# interpolation
# -------------

# using these points ...
px,py,pz = data['coord_x'],data['coord_y'],data['coord_z']
#points = list(itertools.product(px,py,pz))

# ... interp data to these points:
ntheta = 60 
nphi = 31
i = np.arange(ntheta)
j = np.arange(nphi)
theta = np.pi/ntheta*(0.5 + i)
phi = 2.0*np.pi/nphi*(0.5 + j)
th, ph = np.meshgrid(theta, phi,indexing="ij")

sinth = np.sin(th)
costh = np.cos(th)
sinph = np.sin(ph)
cosph = np.cos(ph)

r = 2.5    
x = r * sinth.flatten() * cosph.flatten()
y = r * sinth.flatten() * sinph.flatten()
z = r * costh.flatten()

# test single pt
#pt = 8.011632661654702e-18, 0.13083989060736062, -2.4965738368864345
#pt = 0., 0.13083989060736062, -2.4965738368864345
#x,y,z = np.array([pt[0]]),np.array([pt[1]]),np.array([pt[2]])

xyz = list(zip(x.flatten(),y.flatten(),z.flatten()))

# do it
from scipy.interpolate import griddata,RegularGridInterpolator
ires = {}
for var in data.keys():
    if var in ['time','coord_x','coord_y','coord_z']:
        continue
    #print(data[var].shape)
    #gxxi = griddata(points,data[var].flatten(), (x,y,z), method='linear')
    ires[var] = RegularGridInterpolator((px,py,pz), data[var], method='linear')(xyz)

    
# test interp with exact value
# ----------------------------
from AHFpy.KerrSchild import KerrSchild

ij_indexes = [(1,1),(1,2),(1,3),
              (2,2),(2,3),
              (3,3)]
map_ij = {'xx':(1,1),'xy':(1,2),'xz':(1,3),
          'yy':(2,2),'yz':(2,3),
          'zz':(3,3)}

print("Test interpolation\n Interpolated pts:")
print("---"*20)
print("i  (x,y,z)   radius")
print("\t field   exact   interp   rel.err.")
print("---"*20)
relerr = {}
for var in ires.keys():
    relerr[var] = []
for i in range(len(x)):
    r = np.sqrt(x[i]**2 + y[i]**2 + z[i]**2)

    print("i:{:03d}: ({:+.4e},{:+.4e},{:+.4e}) r={:+.3e}"
          .format(i,x[i],y[i],z[i],r))

    ks = KerrSchild(x = np.array(x[i]),
                    y = np.array(y[i]),
                    z = np.array(z[i]))
    g = ks.Metric()
    K = ks.ExtrinsicCurvature()

    for var in ires.keys():
        
        intrp = ires[var][i]
        if var[0] == 'g':
            exact = g[map_ij[var[1:]]].flatten()[0]
        else:
            exact = K[map_ij[var[1:]]].flatten()[0]
        
        relerr_i = (exact-intrp)/exact
        relerr[var].append(relerr_i)

        print("\t {:<5} {:+.4e} {:+.4e} {:+.2e}"
              .format(var,exact,intrp,relerr_i))

# max err
for var in ires.keys():
    i = np.argmax(np.fabs(relerr[var]))
    m = relerr[var][i]
    print("Max abs. rel.err. of {:<3} is {:+4e} at i = {:03d} : {}".format(var,m,i,xyz[i]))


